"""Classes generating synthetic energy data using a thermo-sensitive model."""

from collections.abc import Callable
from typing import TypedDict

import numpy as np
import pandas as pd

# change the default display options for the doctest
pd.options.display.max_columns = 10
pd.options.display.width = 256


class SynthDDConsumption:
    """Class to generate synthetic thermosensitive energy consumption data.

    The generated consumption is split into a base, a thermosensitive and a residual
    contributions.

    .. note::

        Only one DD can be used at a time to generate the energy consumption.
        For multiple DDs, use the :py:class:`FakeTSConsumption` class.


    Example:
    --------
    >>> synth_consumption = SynthDDConsumption(base_energy=100, ts_slope=0.1,
        noise_std=5)
    >>> synth_consumption.random_consumption(size=5)
                     DD  base  thermosensitive  residual      energy
    2022-11-01  7.212626   100         0.721263 -6.510898   94.210365
    2022-11-02  7.008569   100         0.700857  0.639202  101.340059
    2022-11-03  7.154283   100         0.715428 -1.581213   99.134215
    2022-11-04  0.839383   100         0.083938 -0.084006   99.999932
    2022-11-05  0.259312   100         0.025931 -4.265220   95.760712

    """

    def __init__(
        self,
        base_energy: float,
        ts_slope: float,
        noise_std: float,
        noise_seed: int = 42,
        exp_scale_dd: float = 3.0,
        *,
        clip_negative: bool = True,
    ) -> None:
        """Return a ``FakeDDConsumption`` instance.

        Parameters
        ----------
        base_energy : float
            The value of the averaged non thermosensitive energy consumption.
        ts_slope : float
            The value of the consumption thermosensitivity given in unit of
            energy (same as base) per degree-day.
        noise_std : float
            The standard deviation of the gaussian noise added to the base
            consumption.
        noise_seed : int, optional
            Seed value for the random number generator bound to ``self``.
        exp_scale_dd : float, optional
            Scale (mean) parameter of the exponential distribution used to generate
            fake DD samples. Default is 5°C.
        clip_negative : bool, default : True
            If True, the energy is clipped so that it cannot be below 0.

        """
        self.base_energy = base_energy
        self.ts_slope = ts_slope
        self.noise_std = noise_std
        self._rng = np.random.default_rng(seed=noise_seed)
        self.exp_scale_dd = exp_scale_dd
        self.clip_negative = clip_negative

    def random_dds(
        self,
        size: int = 100,
        start: pd.Timestamp | str = "2022-11-01",
    ) -> pd.Series:
        """Return realistic DD samples.

        Parameters
        ----------
        size : int, default : 100
            Number of days in the sample.
        start : pd.Timestamp or alike, optional
            First day in the generated sample.

        Returns
        -------
        pd.Series :
            A time-series with 1 day period containing randomly
            generated DD values.

        .. warning::

            The date provided in the sample is not consistent
            with the DD values.
            Still, the distribution of the DDs remain correct
            when using large enough number of samples.


        Notes
        -----
        The samples are drawn from an exponential law.

        Example
        -------
        >>> synth_consumption = SynthDDConsumption(base_energy=100, ts_slope=0.1,
            noise_std=5)
        >>> synth_consumption.random_dds(size=5)
        2022-11-01    7.212626
        2022-11-02    7.008569
        2022-11-03    7.154283
        2022-11-04    0.839383
        2022-11-05    0.259312
        Freq: D, Name: DD, dtype: float64

        """
        return pd.Series(
            self._rng.exponential(
                scale=self.exp_scale_dd,
                size=size,
            ),
            index=pd.date_range(start=start, periods=size, freq="1D"),
            name="DD",
        )

    def random_consumption(
        self,
        size: int = 100,
        start: pd.Timestamp | str = "2022-11-01",
    ) -> pd.DataFrame:
        """Return a random decomposed energy consumption.

        Parameters
        ----------
        size : int, default : 100
            Number of days in the sample.
        start : pd.Timestamp or alike, optional
            First day in the generated sample.

        Returns
        -------
        pd.DataFrame :
            The table of fake consumption, as described in :py:meth:`.fake_energy`.


        .. warning::

            The date provided in the sample is not consistent
            with the DD values.
            Still, the distribution of the DDs remain correct
            when using large enough number of samples.

        Example
        -------
        >>> synth_consumption = SynthDDConsumption(base_energy=100, ts_slope=2,
            noise_std=5)
        >>> synth_consumption.random_consumption(size=5)
                         DD  base  thermosensitive  residual      energy
        2022-11-01  7.212626   100        14.425252 -6.510898  107.914354
        2022-11-02  7.008569   100        14.017138  0.639202  114.656340
        2022-11-03  7.154283   100        14.308566 -1.581213  112.727353
        2022-11-04  0.839383   100         1.678766 -0.084006  101.594760
        2022-11-05  0.259312   100         0.518624 -4.265220   96.253405

        """
        return self.fake_energy(self.random_dds(size=size, start=start))

    def fake_energy(
        self,
        dd_samples: pd.Series,
    ) -> pd.DataFrame:
        r"""Return a fake energy consumption for each day in the DD samples.

        Parameters
        ----------
        dd_samples : pd.Series
            A timeseries of DDs to be used to infer the energy consumption.

        Returns
        -------
        pd.DataFrame :
            A table with rows labeled by ``dd_samples`` index and the
            following columns :

            - ``DD``: the ``dd_samples`` series.
            - ``energy``: the energy consumption for each raw in the table.
            - ``thermosensitive``: the value of the thermosensitive energy
              consumption for each period.
            - ``base``: the value of the averaged non-thermosensitive consumption.
              This value is constant across the table, equal to ``self.base_energy``.
            - ``residual``: the energy-noise, i.e. the residual between the
              affine model and the actual energy for each period.

        Notes
        -----
        The energy is generated assuming that the energy consumption satisfies
        the following equation:

        .. math::

            E = \\Theta. DD + E_{base} + \\epsilon

        Where :math:`\\epsilon` is a gaussian, centered, random variable, which
        standard deviation is ``self.noise_std``.

        Example
        -------
        >>> synth_consumption = SynthDDConsumption(base_energy=100, ts_slope=2,
            noise_std=5)
        >>> dds = pd.Series(data=[0,2,12])
        >>> synth_consumption.fake_energy(dds)
           DD  base  thermosensitive  residual      energy
        0    0   100                0  1.523585  101.523585
        1    2   100                4 -5.199921   98.800079
        2   12   100               24  3.752256  127.752256

        """
        fake_data = pd.DataFrame(
            np.empty((dd_samples.shape[0], 5)),
            index=dd_samples.index,
            columns=["DD", "base", "thermosensitive", "residual", "energy"],
        )
        fake_data["DD"] = dd_samples
        fake_data["thermosensitive"] = dd_samples * self.ts_slope
        fake_data["base"] = self.base_energy
        fake_data["residual"] = self._rng.normal(
            loc=0.0,
            scale=self.noise_std,
            size=dd_samples.size,
        )
        assembled_energy = fake_data.loc[
            :,
            ["base", "thermosensitive", "residual"],
        ].sum(axis=1)
        if self.clip_negative:
            assembled_energy = assembled_energy.clip(lower=0.0)
            fake_data["energy"] = assembled_energy
            fake_data["residual"] = fake_data["energy"] - fake_data.loc[
                :,
                ["base", "thermosensitive"],
            ].sum(axis=1)
        else:
            fake_data["energy"] = assembled_energy
        return fake_data

    def measures(
        self,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Return a fake energy consumption  decomposition VS temperature.

        This method is a wrapper around :py:meth:`.random_consumption` to keep the same
        signature as the Handlers.

        See :py:meth:`random_consumption`

        """
        return self.random_consumption(*args, **kwargs)


class SynthTSConsumption:
    """Generates synthetic energy consumption with linear degree-day dependencies.

    A class to generate fake energy consumptions as a function of the
    temperature. Based on :py:class:`FakeDDConsumption`, including both heating
    and cooling domains.

    The generation relies on the assumption of linear DD dependencies in the
    heating and cooling domains.

    Example:
    --------
    >>> synth_consumption = SynthTSConsumption(base_energy=100, ts_heat=2, ts_cool=3,
    ...     noise_std=5)
    >>> synth_consumption.random_consumption(size=5, t_mean=20, t_std=20)
                base  thermosensitive  residual      energy    heating    \
        cooling          T  DD_heating  DD_cooling
        2022-11-01   100        18.283025 -6.510898  111.772127   0.000000  \
        18.283025  26.094342     0.000000     6.094342
        2022-11-02   100        35.599364  0.639202  136.238566  35.599364   \
        0.000000  -0.799682    17.799682     0.000000
        2022-11-03   100        45.027072 -1.581213  143.445859   0.000000  \
        45.027072  35.009024     0.000000    15.009024
        2022-11-04   100        56.433883 -0.084006  156.349877   0.000000  \
        56.433883  38.811294     0.000000    18.811294
        2022-11-05   100        72.041408 -4.265220  167.776188  72.041408   \
        0.000000 -19.020704    36.020704     0.000000

    """

    def __init__(
        self,
        base_energy: float,
        ts_heat: float,
        ts_cool: float,
        t_ref_heat: float = 17,
        t_ref_cool: float = 20,
        noise_std: float = 0,
        noise_seed: int = 42,
    ) -> None:
        """Return a ``SynthTSConsumption`` instance.

        Parameters
        ----------
        base_energy : float
            The value of the averaged non-thermosensitive consumption. Its unit
            can be anything consistent with which of ``ts_heat`` and ``ts_cool``
            when describing an energy.
        ts_heat : float
            The thermosensitivity of the consumption in the heating domain, i.e.
            under ``self.t_ref_heat``. Its has the dimension of one unit of
            ``base_energy`` per degree day.
        ts_cool : float
            The thermosensitivity of the consumption in the cooling domain, i.e.
            over ``self.t_ref_cool``. Same unit as ``ts_heat``.
        t_ref_heat : float, default : 17
            The reference temperature of the heating domain, i.e. the outdoor
            temperature under which the heating is assumed to start.
        t_ref_cool : float, default : 20
            The reference temperature of the cooling domain, i.e. the outdoor
            temperature over which the cooling is assumed to start.
        noise_std : float, default : 0.
            The standard deviation of the gaussian noise added to the affine per
            part model used to generate the energy consumption from the temperature.
        noise_seed : int, default : 42
            A seed for the random noise generator bound to ``self``.

        """
        self.heating = SynthDDConsumption(
            base_energy=0,
            ts_slope=ts_heat,
            noise_std=0,
            clip_negative=False,
            noise_seed=noise_seed,
        )
        self.cooling = SynthDDConsumption(
            base_energy=0,
            ts_slope=ts_cool,
            noise_std=0,
            clip_negative=False,
            noise_seed=noise_seed,
        )
        self.base_energy = base_energy
        self.noise_std = noise_std
        self.t_ref_heat = t_ref_heat
        self.t_ref_cool = t_ref_cool
        self._rng = np.random.default_rng(seed=noise_seed)

    def fake_energy(
        self,
        dd_heating: pd.Series,
        dd_cooling: pd.Series,
        t_samples: pd.Series,
    ) -> pd.DataFrame:
        """Return fake energy data depending on input daily temperatures.

        Parameters
        ----------
        dd_heating : pd.Series
            A series of DD. Usually daily aggregates, depending on the
            scale chosen for the thermosensitivity and base consumption values.
        dd_cooling : pd.Series
            A series of DD. Usually daily aggregates, depending on the
            scale chosen for the thermosensitivity and base consumption values.
        t_samples : pd.Series
            A time series of daily temperatures sampled from a Gaussian distribution,
            representing the temperature values (in °C) for the specified date range.

        Returns
        -------
        pd.DataFrame :
            A table with rows labeled by ``dd_samples`` index and the
            following columns :

            - ``T`` : the ``t_samples`` series.
            - ``energy`` : the energy consumption for each raw in the table.
            - ``thermosensitive`` : the value of the thermosensitive energy
              consumption for each period.
            - ``base`` : the value of the averaged non-thermosensitive consumption.
              This value is constant across the table, equal to ``self.base_energy``.
            - ``residual`` : the energy-noise, i.e. the residual between the
              affine model and the actual energy for each period.

        Notes
        -----
        The fake energy generation relies on two instances of
        :py:class:`SynthDDConsumption`, each one being associated with one of the
        heating and cooling temperature domains which bounds are defined by
        ``self.t_ref_heat`` and ``self.t_ref_cool``, leading to an energy which is
        assumed to be affine per part depending on the temperature:

        - with slope ``-self.ts_heat`` in the heating domain (colder than
          ``self.t_ref_heat``.);
        - with slope ``self.ts_cool`` in the cool domain (warmer than
          ``self.t_ref_cool``.);
        - with slope 0 and constant value ``self.base_energy`` between
          ``self.t_ref_heat`` and ``self.t_ref_cool``.

        Example
        -------
        >>> synth_consumption = SynthTSConsumption(base_energy=100, ts_heat=2,
            ts_cool=0.2)
        >>> dds_cool = pd.Series(data=[0,2,12])
        >>> dds_heat = pd.Series(data=[52,1,0])
        >>> t_samples = pd.Series(data=[10, 15, 20])
        >>> synth_consumption.fake_energy(dds_heat, dds_cool, t_samples)
           base  thermosensitive  residual  energy  heating  cooling   T  DD_heating  \
               DD_cooling
        0   100            104.0       0.0   204.0      104      0.0  10           52 \
            0
        1   100              2.4       0.0   102.4        2      0.4  15            1 \
            2
        2   100              2.4       0.0   102.4        0      2.4  20            0 \
            12

        """
        heating = self.heating.fake_energy(dd_heating)
        cooling = self.cooling.fake_energy(dd_cooling)
        dd_heating = heating.pop("DD")
        dd_cooling = cooling.pop("DD")
        fake_data = heating + cooling
        fake_data["heating"] = heating["thermosensitive"]
        fake_data["cooling"] = cooling["thermosensitive"]
        fake_data["residual"] = self._rng.normal(
            loc=0.0,
            scale=self.noise_std,
            size=dd_cooling.size,
        )
        fake_data["energy"] += fake_data["residual"] + self.base_energy
        fake_data["base"] = self.base_energy
        fake_data["T"] = t_samples
        fake_data["DD_heating"] = dd_heating
        fake_data["DD_cooling"] = dd_cooling
        return fake_data

    def random_dds(
        self,
        t_mean: float = 15,
        t_std: float = 5,
        size: int = 100,
        start: str | pd.Timestamp = "2022-11-01",
        end: str | pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        """Return realistic DD samples.

        Parameters
        ----------
        t_mean : float, default 15
            The average of the gaussian temperature distribution.
        t_std : float, default 5
            The std if the gaussian temperature distribution.
        size : int, optional
            The number of samples to generate. Default is 100.
        start : pd.Timestamp or alike, optional
            The first date of the generated time-series.
            Default is "2022-11-01".
        end : pd.Timestamp or alike, optional
            The last date of the generated time-series.
            Default is None.

        Returns
        -------
        pd.DataFrame :
            A dataframe with 1-day period containing randomly
            generated DD values.

        .. warning::

            The date provided in the sample is not consistent
            with the DD values (think winter in July!)
            Still, the distribution of the DDs remain correct
            when using large enough number of samples.


        Notes
        -----
        The daily mean temperature is drawn from a Normal distribution.
        The DD uses the mean method to compute the DD values.

        Example
        -------
        >>> synth_consumption = SynthTSConsumption(base_energy=100, ts_heat=2,
            ts_cool=0.2)
        >>> synth_consumption.random_dds(size=5, t_mean=20, t_std=20)
                    DD_heating  DD_cooling          T
        2022-11-01     0.000000     6.094342  26.094342
        2022-11-02    17.799682     0.000000  -0.799682
        2022-11-03     0.000000    15.009024  35.009024
        2022-11-04     0.000000    18.811294  38.811294
        2022-11-05    36.020704     0.000000 -19.020704

        """
        index = pd.date_range(start=start, end=end, periods=size, freq="1D")
        size = len(index)
        t_samples = pd.Series(
            self._rng.normal(loc=t_mean, scale=t_std, size=size),
            name="T(°C)",
            index=index,
        )
        fake_dds = pd.concat(
            [
                (self.t_ref_heat - t_samples).clip(lower=0),
                (t_samples - self.t_ref_cool).clip(lower=0),
                t_samples,
            ],
            axis=1,
        )
        fake_dds.columns = pd.Index(["DD_heating", "DD_cooling", "T"])
        return fake_dds

    def random_consumption(
        self,
        size: int = 100,
        t_mean: float = 15,
        t_std: float = 5,
        start: pd.Timestamp | str = "2022-11-01",
        end: pd.Timestamp | str | None = None,
    ) -> pd.DataFrame:
        """Return a fake energy consumption decomposition VS temperature.

        The input temperatures are generated using a gaussian distribution.

        Parameters
        ----------
        size : int, optional
            The number of samples to generate. Default is 100.
        t_mean : float, default 15
            The average of the gaussian temperature distribution.
        t_std : float, default 5
            The std if the gaussian temperature distribution.
        start : pd.Timestamp or alike, optional
            The first date of the generated time-series.
            Default is "2022-11-01".
        end : pd.Timestamp or alike, optional
            The last date of the generated time-series.
            Default is None.

        Returns
        -------
        pd.DataFrame :
            A table of randomly generated energy consumption as a function of the
            temperature. See :py:meth:`.fake_energy`.

        """
        dd_samples = self.random_dds(
            size=size,
            t_mean=t_mean,
            t_std=t_std,
            start=start,
            end=end,
        )
        return self.fake_energy(
            dd_samples["DD_heating"],
            dd_samples["DD_cooling"],
            dd_samples["T"],
        )

    def measures(
        self,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Return a fake energy consumption  decomposition VS temperature.

        This method is a wrapper around :py:meth:`.random_consumption` to keep the same
        signature as the Handlers.

        """
        return self.random_consumption(*args, **kwargs)


class DateSynthTSConsumption(SynthTSConsumption):
    """Generates synthetic energy data based on date-driven temperature variations.

    A class to generate fake energy consumptions as a function of the
    temperature. Based on :py:class:`SynthTSConsumption`, including both heating
    and cooling domains.

    The generation relies on the assumption of linear DD dependencies in the
    heating and cooling domains.

    This class extends the :py:class:`SynthTSConsumption` to generate a realistic
    temperature as function of the date.
    """

    def __init__(
        self,
        base_energy: float,
        ts_heat: float,
        ts_cool: float,
        t_ref_heat: float = 17,
        t_ref_cool: float = 20,
        noise_std: float = 0,
        noise_seed: int = 42,
        temperature_amplitude_year: float = 9.36,
        temperature_period_year: float = 2 * np.pi / 364.991,
        temperature_phase_year: int = 13,
    ) -> None:
        """Return a ``DateSynthTSConsumption`` instance.

        Parameters
        ----------
        base_energy : float
            The value of the averaged non-thermosensitive consumption. Its unit
            can be anything consistent with which of ``ts_heat`` and ``ts_cool``
            when describing an energy.
        ts_heat : float
            The thermosensitivity of the consumption in the heating domain, i.e.
            under ``self.t_ref_heat``. Its has the dimension of one unit of
            ``base_energy`` per degree day.
        ts_cool : float
            The thermosensitivity of the consumption in the cooling domain, i.e.
            over ``self.t_ref_cool``. Same unit as ``ts_heat``.
        t_ref_heat : float, default : 17
            The reference temperature of the heating domain, i.e. the outdoor
            temperature under which the heating is assumed to start.
        t_ref_cool : float, default : 20
            The reference temperature of the cooling domain, i.e. the outdoor
            temperature over which the cooling is assumed to start.
        noise_std : float, default : 0.
            The standard deviation of the gaussian noise added to the affine per
            part model used to generate the energy consumption from the temperature.
        noise_seed : int, default : 42
            A seed for the random noise generator bound to ``self``.
        temperature_amplitude_year : float, optional
            The temperature amplitude over the year, by default 9.36.
            Think of it as the difference between the hottest and coldest days.
        temperature_period_year : float, optional
            The period of the year, by default 2*np.pi/364.991
            Should be close to 2*np.pi/365.
        temperature_phase_year : int, optional
            The phase, in days, by default 13

        """
        super().__init__(
            base_energy=base_energy,
            ts_heat=ts_heat,
            ts_cool=ts_cool,
            t_ref_heat=t_ref_heat,
            t_ref_cool=t_ref_cool,
            noise_std=noise_std,
            noise_seed=noise_seed,
        )
        self.temperature_amplitude_year = temperature_amplitude_year
        self.temperature_period_year = temperature_period_year
        self.temperature_phase_year = temperature_phase_year

    def synthetic_temperature(
        self,
        t_mean: float = 14,
        t_std: float = 5,
        size: int = 100,
        start: pd.Timestamp | str = "2022-11-01",
        end: pd.Timestamp | str | None = None,
    ) -> pd.Series:
        """Return a synthetic temperature time-series.

        .. note::

            The temperature is generated using a sinusoidal model with a gaussian
            noise added.
            The parameters of the sinusoidal model are the class attributes.

            The noise model should be improved in the future, as the temperature
            fluctuations present actually several frequencies.

        Parameters
        ----------
        t_mean : float, default 15
            The average of the gaussian temperature distribution.
        t_std : float
            The standard deviation of the gaussian noise added to the temperature
            base.
        size : int, optional
            The number of samples to generate. Default is 100.
        start : pd.Timestamp or alike, optional
            The first date of the generated time-series.
            Default is "2022-11-01".
        end : pd.Timestamp or alike, optional
            The last date of the generated time-series.
            Default is None.

        Returns
        -------
        pd.Series :
            A time-series of synthetic temperatures.

        Example
        -------
        >>> synth_consumption = DateTimeSynthTSConsumption(base_energy=100, ts_heat=2,
            ts_cool=0.2)
        >>> synth_consumption.synthetic_temperature(size=5)
        2022-11-01    13.187131
        2022-11-02     6.307951
        2022-11-03    15.105192
        2022-11-04    15.901608
        2022-11-05     1.290287
        Freq: D, Name: T(°C), dtype: float64

        Notes
        -----
        From the 3 parameters `size`, `start` and `end`, only two must be given.

        """
        index = pd.date_range(
            start=start,
            periods=size,
            end=end,
            freq="1D",
        )
        julian_date = index.to_julian_date()
        temperature_base = t_mean + self.temperature_amplitude_year * np.sin(
            self.temperature_period_year * (julian_date - self.temperature_phase_year),
        )
        return pd.Series(
            self._rng.normal(loc=temperature_base, scale=t_std, size=len(index)),
            index=index,
            name="T(°C)",
        )

    def random_dds(
        self,
        t_mean: float = 15,
        t_std: float = 5,
        size: int = 100,
        start: str | pd.Timestamp = "2022-11-01",
        end: str | pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        r"""Return realistic DD samples.

        Parameters
        ----------
        t_mean : float, default 15
            The average of the gaussian temperature distribution.
        t_std : float, optional
            The standard deviation of the gaussian noise added to the temperature
            base. Default is 5.
        size : int, optional
            The number of samples to generate. Default is 100.
        start : pd.Timestamp or alike, optional
            The first date of the generated time-series.
            Default is "2022-11-01".
        end : pd.Timestamp or alike, optional
            The last date of the generated time-series.
            Default is None.

        Returns
        -------
        pd.DataFrame :
            A DataFrame with the following columns :
            - ``DD_heating`` : the heating DD values.
            - ``DD_cooling`` : the cooling DD values.
            - ``T`` : the temperature.

        Notes
        -----
        The DDs are computed using the mean temperature of the day as

        .. math::

                DD = \\max(0, T_{mean} - T_{ref})

        Hence, it is really just the same as the mean temperature of the day.

        From the 3 parameter `size`, `start` and `end`, only two must be
        given.

        Example
        -------
        >>> synth_consumption = DateTimeSynthTSConsumption(base_energy=100, ts_heat=2,
            ts_cool=0.2)
        >>> synth_consumption.random_dds(size=5, t_std=25)
                    DD_heating  DD_cooling          T
        2022-11-01     0.000000     0.000000  19.281473
        2022-11-02    31.491731     0.000000 -14.491731
        2022-11-03     0.000000    10.114216  30.114216
        2022-11-04     0.000000    14.712902  34.712902
        2022-11-05    54.730417     0.000000 -37.730417

        """
        t_samples = self.synthetic_temperature(
            t_mean=t_mean,
            t_std=t_std,
            size=size,
            start=start,
            end=end,
        )
        dd_heating = (self.t_ref_heat - t_samples).clip(lower=0)
        dd_cooling = (t_samples - self.t_ref_cool).clip(lower=0)
        data = pd.concat([dd_heating, dd_cooling, t_samples], axis=1)
        data.columns = pd.Index(["DD_heating", "DD_cooling", "T"])
        return data


class TSParameters(TypedDict):
    """TypedDict for specifying thermosensitive model parameters.

    The ``TSParameters`` class defines the structure of parameters used to generate
    synthetic energy data with a thermo-sensitive model. It includes fields for base
    energy consumption and thermosensitivity in both heating and cooling domains, as
    well as noise standard deviation.

    Attributes
    ----------
    base_energy : float
        The base energy consumption that is independent of temperature variations.
    ts_heat : float
        The slope that represents how much the energy consumption increases per
        degree day below the reference heating temperature (thermosensitivity for
        heating).
    ts_cool : float
        The slope that represents how much the energy consumption increases per
        degree day above the reference cooling temperature (thermosensitivity for
        cooling).
    noise_std : float
        The standard deviation of the Gaussian noise added to simulate randomness
        in the energy consumption.

    """


class CategorySynthTSConsumption(DateSynthTSConsumption):
    """Generates synthetic energy consumption with temperature-based categorization.

    A class to generate fake energy consumptions as a function of the
    temperature. Based on :py:class:`DateSynthTSConsumption`, including both heating
    and cooling domains.

    Add the possibility to generate different categories of energy consumption
    with a function that categorize the periods.

    .. note::

        The categories are labeled by a function. The base temperatures
        (``t_ref_heat`` and ``t_ref_cool``) are the same for all categories.


    For an example, see the :py:class:`WeekEndSynthTSConsumption` class
    that implement the concept of week-end and week days.
    """

    def __init__(
        self,
        parameters: list[TSParameters],
        t_ref_heat: float,
        t_ref_cool: float,
        list_categories: list,
        category_func: Callable,
        noise_seed: int = 42,
        temperature_amplitude_year: float = 9.36,
        temperature_period_year: float = 2 * np.pi / 364.991,
        temperature_phase_year: int = 13,
    ) -> None:
        """Return a ``CategorySynthTSConsumption`` instance.

        Parameters
        ----------
        parameters : list[TSParameters]
            A list of dictionaries containing the parameters for each category.
        t_ref_heat : float
            The reference temperature of the heating domain, i.e. the outdoor
            temperature under which the heating is assumed to start.
        t_ref_cool : float
            The reference temperature of the cooling domain, i.e. the outdoor
            temperature over which the cooling is assumed to start.
        list_categories : list
            The list of the different categories labels.
        category_func : Callable
            A function that takes a pd.Series with DateTimeIndex as input and
            return a pd.Series with the categories labels.
        noise_seed : int, default : 42
            A seed for the random noise generator bound to ``self``.
        temperature_amplitude_year : float, optional
            The temperature amplitude over the year, by default 9.36.
            Think of it as the difference between the hottest and coldest days.
        temperature_period_year : float, optional
            The period of the year, by default 2*np.pi/364.991
            Should be close to 2*np.pi/365.
        temperature_phase_year : int, optional
            The phase, in days, by default 13

        Notes
        -----
        The number of categories in ``list_categories`` must match the number of
        ``parameters``. In assertion, the order of the categories must match the order
        of the parameters.

        """
        self.list_of_synths = [
            DateSynthTSConsumption(
                **param,
                t_ref_heat=t_ref_heat,
                t_ref_cool=t_ref_cool,
                noise_seed=noise_seed,
                temperature_amplitude_year=temperature_amplitude_year,
                temperature_period_year=temperature_period_year,
                temperature_phase_year=temperature_phase_year,
            )
            for param in parameters
        ]
        self.t_ref_cool = t_ref_cool
        self.t_ref_heat = t_ref_heat
        self.temperature_amplitude_year = temperature_amplitude_year
        self.temperature_period_year = temperature_period_year
        self.temperature_phase_year = temperature_phase_year
        self._rng = np.random.default_rng(seed=noise_seed)
        self.list_categories = list_categories
        self.category_func = category_func
        if len(self.list_of_synths) != len(self.list_categories):
            err = "The number of categories must match the number of synthetizers"
            raise ValueError(err)

    def fake_energy(
        self,
        dd_heating: pd.Series,
        dd_cooling: pd.Series,
        t_samples: pd.Series,
    ) -> pd.DataFrame:
        """Return fake energy consumption based on input daily temperatures.

        Parameters
        ----------
        dd_heating : pd.Series
            A series of DD. Usually daily aggregates, depending on the
            scale chosen for the thermosensitivity and base consumption values.
        dd_cooling : pd.Series
            A series of DD. Usually daily aggregates, depending on the
            scale chosen for the thermosensitivity and base consumption values.
        t_samples : pd.Series
            A time series of daily temperatures sampled from a Gaussian distribution,
            representing the temperature values (in °C) for the specified date range.

        Returns
        -------
        pd.DataFrame :
            A table with rows labeled by ``dd_samples`` index and the
            following columns :

            - ``T`` : the ``t_samples`` series.
            - ``energy`` : the energy consumption for each row in the table.
            - ``thermosensitive`` : the value of the thermosensitive energy
              consumption for each period.
            - ``base`` : the value of the averaged non-thermosensitive consumption.
              This value is constant across the table, equal to ``self.base_energy``.
            - ``residual`` : the energy-noise, i.e. the residual between the
              affine model and the actual energy for each period.

        Notes
        -----
        The fake energy generation relies on two instances of
        :py:class:`SynthDDConsumption`, each one being associated with one of
        the heating and cooling temperature domains which bounds are defined by
        ``self.t_ref_heat`` and ``self.t_ref_cool``, leading to an energy which
        is assumed to be affine per part depending on the temperature:

        - with slope ``-self.ts_heat`` in the heating domain (colder than
          ``self.t_ref_heat``);
        - with slope ``self.ts_cool`` in the cool domain (warmer than
          ``self.t_ref_cool``);
        - with slope 0 and constant value ``self.base_energy`` between
          ``self.t_ref_heat`` and ``self.t_ref_cool``.

        See :py:class:`SynthTSConsumption` for more details.

        Example
        -------
        >>> synth_consumption = SynthTSConsumption(base_energy=100, ts_heat=2,
            ts_cool=0.2)
        >>> dds_cool = pd.Series(data=[0,2,12])
        >>> dds_heat = pd.Series(data=[52,1,0])
        >>> t_samples = pd.Series(data=[10, 15, 20])
        >>> synth_consumption.fake_energy(dds_heat, dds_cool, t_samples)
        base  thermosensitive  residual  energy  heating  cooling   T  DD_heating \
            DD_cooling
        0   100            104.0       0.0   204.0      104      0.0  10           52 \
            0
        1   100              2.4       0.0   102.4        2      0.4  15            1 \
            2
        2   100              2.4       0.0   102.4        0      2.4  20            0 \
            12

        """
        categories_series = self.category_func(t_samples)
        list_of_fake_data = []
        for category, synth in zip(
            self.list_categories,
            self.list_of_synths,
            strict=False,
        ):
            mask = categories_series == category
            fake_data = synth.fake_energy(
                dd_heating[mask],
                dd_cooling[mask],
                t_samples[mask],
            )
            fake_data["category"] = category
            list_of_fake_data.append(fake_data)
        return pd.concat(list_of_fake_data, axis=0).sort_index()


class WeekEndSynthTSConsumption(CategorySynthTSConsumption):
    """Generates synthetic energy consumption categorized by weekdays and weekends.

    A class to generate fake energy consumptions as a function of the
    temperature with two categories: week days and weekends.

    Based on :py:class:`CategorySynthTSConsumption`.
    """

    def __init__(
        self,
        parameters: list[TSParameters],
        t_ref_heat: float,
        t_ref_cool: float,
        noise_seed: int = 42,
        temperature_amplitude_year: float = 9.36,
        temperature_period_year: float = 2 * np.pi / 364.991,
        temperature_phase_year: int = 13,
    ) -> None:
        """Return a ``WeekEndSynthTSConsumption`` instance.

        Parameters
        ----------
        parameters : list[TSParameters]
            A list of dictionaries containing the parameters for the two categories:

            1. The first dictionary is for the week days.
            2. The second dictionary is for the weekends.
        t_ref_heat : float, default : 17
            The reference temperature of the heating domain, i.e. the outdoor
            temperature under which the heating is assumed to start.
        t_ref_cool : float, default : 20
            The reference temperature of the cooling domain, i.e. the outdoor
            temperature over which the cooling is assumed to start.
        noise_seed : int, default : 42
            A seed for the random noise generator bound to ``self``.
        temperature_amplitude_year : float, optional
            The temperature amplitude over the year, by default 9.36.
            Think of it as the difference between the hottest and coldest days.
        temperature_period_year : float, optional
            The period of the year, by default 2*np.pi/364.991
            Should be close to 2*np.pi/365.
        temperature_phase_year : int, optional
            The phase, in days, by default 13.

        Example
        -------
        >>> parameters = [
        ...     {"base_energy": 100, "ts_heat": 2, "ts_cool": 0.2, "noise_std": 5},
        ...     {"base_energy": 100, "ts_heat": 1, "ts_cool": 0.1, "noise_std": 5},
        ... ]
        >>> synth_consumption = WeekEndSynthTSConsumption(parameters, t_ref_heat=17,
            t_ref_cool=20)

        """
        list_categories = ["weekend", "weekday"]

        def category_func(
            t_samples: pd.Series,
        ) -> np.ndarray:
            friday_weekday = 5
            index = t_samples.index
            if not isinstance(index, pd.DatetimeIndex):
                err = "Index must be a DatetimeIndex"
                raise TypeError(err)
            return np.where(
                index.dayofweek < friday_weekday,
                "weekday",
                "weekend",
            )

        super().__init__(
            parameters=parameters,
            t_ref_heat=t_ref_heat,
            t_ref_cool=t_ref_cool,
            noise_seed=noise_seed,
            temperature_amplitude_year=temperature_amplitude_year,
            temperature_period_year=temperature_period_year,
            temperature_phase_year=temperature_phase_year,
            list_categories=list_categories,
            category_func=category_func,
        )



"""Defines custom errors for degree-days."""

from .base import EATError


class EATInvalidDegreeDaysMethodError(EATError):
    """The provided method to compute degree-days doesn't exist."""

    def __init__(
        self,
        method_name: str,
        valid_methods: list[str],
    ) -> None:
        """Initialize the exception with invalid and valid method names.

        Parameters
        ----------
        method_name : str
            The name of the method that was provided and found to be invalid.
        valid_methods : list[str]
            A list of valid method names that can be used for computing degree-days.

        Raises
        ------
        EATInvalidDegreeDaysMethod
            If an invalid method name is used for computing degree-days.

        Examples
        --------
        >>> raise EATInvalidDegreeDaysMethod("unknown_method", ["min_max", "mean"])
        EATInvalidDegreeDaysMethod: The error 'unknown_method' doesn't exist.
        Possibilities are min_max, mean.

        """
        self.method_name = method_name
        self.valid_methods = valid_methods
        message = (
            f"The degree-days calculation method '{method_name}' doesn't exist. "
            f"Possibilities are {', '.join(valid_methods)}."
        )
        super().__init__(message)


class EATInvalidDegreeDaysError(EATError):
    """The provided type of degree-days doesn't exist."""

    def __init__(
        self,
        dd_type: str,
        valid_dd_types: list[str],
    ) -> None:
        """Initialize the exception with invalid and valid degree-days types.

        Parameters
        ----------
        dd_type : str
            The name of the degree-day type that was provided and found to be invalid.
        valid_dd_types : list[str]
            A list of valid degree-days names.

        Raises
        ------
        EATInvalidDegreeDaysMethod
            If an invalid method name is used for computing degree-days.

        Examples
        --------
        >>> raise EATInvalidDegreeDaysError("unknown_dd_type", ["heating", "cooling"])
        EATInvalidDegreeDaysMethod: The error 'unknown_dd_type' doesn't exist.
        Possibilities are heating, cooling.

        """
        self.dd_type = dd_type
        self.valid_dd_types = valid_dd_types
        message = (
            f"The degree-days type '{dd_type}' doesn't exist. "
            f"Possibilities are {', '.join(valid_dd_types)}."
        )
        super().__init__(message)



"""Filter history data based on day of week."""

import pandas as pd


def weekdays_only(
    history: pd.DataFrame,
) -> pd.DataFrame:
    """Return the weekdays in history.

    Parameters
    ----------
    history : pd.DataFrame
        History data to be filtered. It is expected that the data is time-indexed,
        with monotonic-increasing labels.

    Returns
    -------
    pd.DataFrame :
        Input history from which all weekends entry have been removed.

    """
    saturday_day_of_week = 5
    return history.loc[history.index.day_of_week < saturday_day_of_week]


def weekends_only(
    history: pd.DataFrame,
) -> pd.DataFrame:
    """Return the weekends in history.

    Parameters
    ----------
    history : pd.DataFrame
        History data to be filtered. It is expected that the data is time-indexed,
        with monotonic-increasing labels.

    Returns
    -------
    pd.DataFrame :
        Input history from which all weekdays entry have been removed.

    """
    friday_day_of_week = 4
    return history.loc[history.index.day_of_week > friday_day_of_week]


def same_day_only(
    history: pd.DataFrame,
    date: pd.Timestamp | None = None,
) -> pd.DataFrame:
    """Return the same days of week in history n_days before date.

    Parameters
    ----------
    history : pd.DataFrame
        History data to be filtered. It is expected that the data is time-indexed,
        with monotonic-increasing labels.
    date : pd.Timestamp, optional
        Reference for which the day-of-week should be the same in history.
        In case ``None`` is passed, the day after the one of the last entry
        in history is used.

    Returns
    -------
    pd.DataFrame :
        Input history in which only the entries for which the day in the week is
        the same as ``date`` have been conserved.

    """
    if history.empty:
        return history
    if date is None:
        date = (history.index[-1].floor("D")) + pd.Timedelta("1D")
    else:
        pass
    return history.loc[history.index.day_of_week == date.day_of_week]



"""Define classes that compute different kind of profiles."""

from . import (
    preprocessing,
    thresholds,
)
from .localization import (
    LocalizedMeanProfile,
    LocalizedRollingProfile,
    LocalizedRollingQuantileProfile,
)
from .mean_profile import MeanProfile
from .rolling_profile import (
    RollingProfile,
    RollingQuantileProfile,
)



"""The module defines rolling aggregations from history.

The two following sections explain the main steps of the
computations :

1. the Pivot phase
2. the Rolling phase

1. The Pivot Phase
------------------
The first step is to aggregates the different days
of the history align by time.

.. image:: /_static/illustration_rollingProfile_1.png
  :width: 550
  :alt: the Pivot phase

To do this step, we need the timestamps to be aligned.
For instance, the time should be rounded to the minute,
as trailing microseconds could impact the pivot.

2. The Rolling agg Phase
------------------------

The second step is to convert the 2D array into a 1D
array with a rolling window aggregation.

.. image:: /_static/illustration_rollingProfile_2.png
  :width: 250
  :alt: the rolling agg phase

In contrast with most rolling window implementations, here
the window rolls along the time-of-day (wall clock time), but not
along the days: all the history data which time-of-day falls within
the window around the target time-of-day is included in the aggregation.
E.g., with a 60min centered window, the data at time t in the returned
profile is the aggregation of all the data in history which time-of-day
is 30min before or after t, whatever the date.

The window size is a duration (like 60min), that can manage
missing data via ``pandas``.

.. note::

    This look-up using ``pandas`` slices is significantly slow.

    A solution to improve the performance of the rolling profile
    is to make sure that all the expected rows are present (filled with NaNs if
    necessary).

    Using this pre-formatting step, we can use ``numpy`` to slice
    using row indexes.

"""

from collections.abc import Callable
from functools import partial

import numpy as np
import pandas as pd

from .mean_profile import MeanProfile


class RollingProfile:
    """Compute a profile by aggregating the history on time-periods rolling windows."""

    def __init__(
        self,
        window: int | str | pd.Timedelta,
        aggregation: Callable[[np.ndarray], float],
        as_mean_offset: bool | None = None,
        offset_factor: float = 1.0,
        column_name: str | int = "value",
    ) -> None:
        """Create a Rolling Agg Profile.

        Parameters
        ----------
        window : int, str, pd.Timedelta
            The rolling window size. Any value which can be passed as argument
            for a `pandas` rolling operation is valid.
            The profile will aggregated the values based on time-of-day for any
            date in history in a window of the chosen size, and associate the
            resulting value to the time-of-day of the *center* of the window in the
            produced profile.
        aggregation : function
            An aggregation function which can work on a 1D numpy array. It is applied to
            a raveled 2D horizontal slice of the pivoted history.
        as_mean_offset : bool, optional
            If True, the resulting profile is multiplied by ``offset_factor`` and
            added to the mean profile obtained with the passed data. Default is False.
        offset_factor : float, optional
            The offset factor, see above. By default 1.
        column_name : str | int, Optional
            The name of the column to process.

        """
        if as_mean_offset is None:
            as_mean_offset = False
        self.window = window
        self.agg = aggregation
        self.as_mean_offset = as_mean_offset
        self.column_name = column_name
        if self.as_mean_offset:
            self.reference = MeanProfile()
            self.offset_factor = offset_factor

    def compute(
        self,
        history: pd.DataFrame,
        time: pd.Timestamp,
        **kwargs,
    ) -> pd.Series:
        """Compute the rolling aggregation profile.

        Parameters
        ----------
        history : pd.DataFrame
            The timeseries of history data, with at least a ``self.column_name``
            column containing the values to be used to create the profile.
            The data should be sampled homogeneously, such that measures are repeated
            at the same time-of-day for every day in the history.
        time : pd.Timestamp
            The time at which the computed profile should start.

        Returns
        -------
        pd.Series
            The computed profile, with same sampling as the history.

        """
        df_day_by_time = self.daily_pivot(history)
        profile = self.windowed_rolling_agg(df_day_by_time)
        profile.index += time
        profile.index.name = history.index.name
        if self.as_mean_offset:
            ref = self.reference.compute(history, time, **kwargs)
            profile = ref.loc[:, [self.column_name]] + self.offset_factor * profile
        return profile

    def daily_pivot(
        self,
        history: pd.DataFrame,
    ) -> pd.DataFrame:
        """Transform history data into a table with time-of-day rows and date columns.

        This function reshapes the historical timeseries data into a pivoted format
        where each column represents a day, and each row corresponds to a specific
        time-of-day. This allows for time-of-day-based aggregations over multiple days.

        Parameters
        ----------
        history : pd.DataFrame
            A DataFrame containing the history data, with a DateTimeIndex. It must
            include the column specified by ``self.column_name``, which represents
            the values to be processed.

        Returns
        -------
        pd.DataFrame
            A pivoted DataFrame where:
            - Each column corresponds to a specific day from the history.
            - Each row represents a specific time-of-day.
            - Values are from the column specified by ``self.column_name``.

        Notes
        -----
        This function handles Daylight Saving Time (DST) changes. Specifically:
          - For winter DST transitions (25-hour days), any times beyond 24 hours
            (e.g., the last hour of the day) are dropped to ensure consistency.
          - Duplicated timestamps due to DST transitions are resolved by keeping the
            first occurrence.

        - Times should be rounded to a consistent resolution (e.g., seconds or minutes)
          to avoid pivoting errors due to small differences in time values.


        .. warning::

            Winter DST creates days lasting 25h, which would lead to 25h profiles.
            This function drops any moment in the day which is more than 24h after
            midnight this day, i.e. drops the last hour of the DST.
            This is not perfect but this is considered satisfactory to begin with.


        """
        history = history.copy()
        history["time"] = history.index - history.index.floor("D")
        history["day"] = history.index.date
        try:
            df_day_by_time = pd.pivot_table(
                history,
                index="time",
                columns=["day"],
                values=[self.column_name],
            )
        except ValueError:
            # Happens on winter DST and time-naive data when the same time happens twice
            df_day_by_time = pd.pivot_table(
                history.loc[~history.index.duplicated(keep="first")],
                index="time",
                columns=["day"],
                values=[self.column_name],
            )
        # Deal with winter DST and time-localized data
        df_day_by_time = df_day_by_time.drop(
            labels=df_day_by_time.index[df_day_by_time.index >= pd.Timedelta("1D")],
        )
        return df_day_by_time[self.column_name]

    def windowed_rolling_agg(
        self,
        pivoted_history: pd.DataFrame,
    ) -> pd.DataFrame:
        """Return rolling aggregation over date/time-of-day-window in a pivoted history.

        Parameters
        ----------
        pivoted_history: pd.DataFrame
            The multi column dataframe on which to compute the aggregation. The
            dataframe columns are series of values (usually a date for each column) and
            the rows are times which can be rolled by a duration window.


        Returns
        -------
        aggregated: pd.Dataframe
            The computed aggregation as a table with
            one column named ``self.column_name``.


        .. note::

            Using the setup allows to be sure that the dataframe ``pivoted_history``
            upon which the agg is computed is the desired one (local scope of
            ``custom_agg`` instead of global scope).

        """

        def custom_agg(
            subseries: pd.Series,
        ) -> float:
            """Compute the ``agg`` function on ``pivoted_history`` using subseries.

            For performance issue, slicing and ``ravel`` are used.
            """
            start, end = subseries.index[0], subseries.index[-1]
            values = pivoted_history.loc[start:end, :].to_numpy().ravel()
            return self.agg(values)

        aggregated = (
            pivoted_history.iloc[:, 0]  # select the first column, as we only roll once.
            .rolling(self.window, center=True)
            .apply(custom_agg, raw=False)  # we need the index passed.
        )
        aggregated.name = self.column_name
        return pd.DataFrame(data=aggregated)


class RollingQuantileProfile(RollingProfile):
    """A profile computed by a Quantile of the history on rolling windows."""

    def __init__(
        self,
        window: int | str | pd.Timedelta,
        threshold_quantile: float,
        as_mean_offset: bool | None = None,
        offset_factor: float = 1.0,
        column_name: str | int = "value",
    ) -> None:
        """Create a Rolling Agg Profile.

        Parameters
        ----------
        window : int, str, pd.Timedelta
            The rolling window size. Any value which can be passed as argument
            for a `pandas` rolling operation is valid.
            The profile will aggregated the values based on time-of-day for any
            date in history in a window of the chosen size, and associate the
            resulting value to the time-of-day of the *center* of the window in the
            produced profile.
        threshold_quantile : float
            The quantile value to compute, in [0:1]
        as_mean_offset : bool, optional
            If True, the resulting profile is multiplied by ``offset_factor`` and
            added to the mean profile obtained with the passed data. Default is False.
        offset_factor : float, optional
            The offset factor, see above. By default 1.
        column_name : str | int, Optional
            The name of the column to process.

        """
        if as_mean_offset is None:
            as_mean_offset = False
        aggregation = partial(np.quantile, q=threshold_quantile)
        super().__init__(
            window,
            aggregation,
            as_mean_offset,
            offset_factor,
            column_name,
        )



import numpy as np
import pandas as pd

DAY_DEFAULT = pd.Timestamp("2020-03-05")

np.random.seed(42)


def example_volume_one_day(day_start=None, total_volume=1.0):
    """Return an example series of "volume" on a day.

    "volume" means any quantity which follows a conservation law.
    The total volume sums to 1(arbitrary unit) so that it can easily be scaled to
    match a target daily volume. The time location of the consumption as well as
    proportions is summarized in the following diagram::

         0  1  2  3  4  5  6  7  8  9  10 11 12 13  14 15 16 17 18 19 20 21 22 23
         Consumption
         |__|__|__|__|__|__|_3|__|._|__|__|__|__|_1.|__|__|__|__|__|__|_2|_3|__|__|

    Parameters
    ----------
    day_start : pd.Timestamp
        Date for which the data are generated. The default is None in case
        data are generated for 2020-03-05.
    total_volume : float
        Total volume consumed during the example day (used to scale the series
        represented above). Default is 1 , meaning that the series returned by
        the function sums up to 1.

    Returns
    -------
    day_example : pd.Series
        An example consumption of an equivalent volume with 30min timestep.
        The series is sampled such that the value with label ``ti`` is the
        volume associated to the interval ``[ti, ti=1[``.

    """
    day_start = day_start or DAY_DEFAULT
    begin = pd.Timestamp(day_start).floor("D")
    index_day = pd.date_range(begin, periods=48, freq="30min")
    day_example = pd.Series(np.zeros(index_day.size), index=index_day)
    day_example.iloc[13] = 0.3  # 6h30 - 30%
    day_example.iloc[16] = 0.05  # 8h - 5%
    day_example.iloc[27] = 0.15  # 13h30 - 15%
    day_example.iloc[41] = 0.2  # 20h30 - 20%
    day_example.iloc[43] = 0.3  # 21h30 - 30%
    return day_example * total_volume



"""Defines all custom errors used in |eat|."""

from .base import (
    EATEmptyDataError,
    EATError,
)
from .invalid_timeseries import (
    EATInvalidTimeseriesError,
    EATInvalidTimestepDurationError,
    EATUndefinedTimestepError,
)
from .resampling import (
    EATEmptySourceError,
    EATEmptyTargetsError,
    EATResamplingError,
)



"""Define the logger for `energy_analysis_toolbox`."""

import inspect
import logging
import sys


def init_logging(
    base_logger_name: str = "energy_analysis_toolbox",
    level: int | str = "INFO",
    min_logger_level_stdout: int | str = "DEBUG",
) -> logging.Logger:
    """Return the project logger.

    Parameters
    ----------
    base_logger_name : str, optional
        The name of the logger. Default is ``energy_analysis_toolbox``.
        If an instance already exists, a number will automatically be
        added to make it unique. That means that if the base_logger_name
        is "cow", the first logger will be called "cow", the second "cow2", ...
    level : str or int, optional
        The level of the logger. Default is ``INFO``.
    min_logger_level_stdout: str, int, optional
        Minimum logger level below which no message is transferred to stdout
        (i.e. not printed). Default is ``"DEBUG"``.

    Returns
    -------
    Logger
        The logger object.


    .. note::

        The usual use is the following:

        .. code-block:: python

            log = init_logging()
            log.debug("a debug message")
            log.info("an info")


    """
    logger_name = _unique_logger_name(base_logger_name)
    log = logging.getLogger(logger_name)
    if not log.hasHandlers():
        log.setLevel(level)
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(min_logger_level_stdout)
        logger_max_length = len(base_logger_name) + 2
        message_fmt = f"%(name)-{logger_max_length}s :: %(levelname)-8s :: %(message)s"
        formatter = MultiLineFormatter(message_fmt)
        ch.setFormatter(formatter)
        log.addHandler(ch)
    return log


def _unique_logger_name(base_logger_name: str) -> str:
    """Generate a unique logger name by appending a counter to the base logger name.

    This function checks if a logger with the specified base name already has
    handlers attached. If it does, the function appends a counter (starting from
    1) to the base logger name, incrementing the counter until it finds a
    logger name that does not already exist. This ensures that the returned
    logger name is unique and can be safely used to create a new logger.

    Parameters
    ----------
    base_logger_name : str
        The base name for the logger. This name will be used as the starting point
        for generating the unique logger name.

    Returns
    -------
    str
        A unique logger name derived from the base logger name. If the base name
        is already unique, it returns the base name itself; otherwise, it returns
        a name in the format `{base_logger_name}_{counter}` where `counter`
        is the smallest integer that makes the name unique.

    Example
    -------
    >>> logger_name = _unique_logger_name("my_logger")
    >>> print(logger_name)
    'my_logger_1'  # If 'my_logger' already exists with handlers

    Notes
    -----
    This function is useful for dynamically creating loggers when multiple
    instances of a class or module might be initialized, ensuring each has a
    distinct logger to avoid conflicts in logging output.

    """
    counter = 1
    logger_name = base_logger_name
    max_counter = 10  # to avoid infinite loops
    while logging.getLogger(logger_name).hasHandlers() or counter < max_counter:
        logger_name = f"{base_logger_name}_{counter}"
        counter += 1
    return logger_name


class MultiLineFormatter(logging.Formatter):
    """Multi-line formatter."""

    def get_header_length(self, record: logging.LogRecord) -> int:
        """Calculate the header length of the given log record.

        This method formats a `LogRecord` without the message and calculates the length
        of the formatted string. The header includes information such as the logger
        name, log level, timestamp, and any other metadata included in the log format.

        Parameters
        ----------
        record : logging.LogRecord
            The log record for which the header length is to be calculated. This record
            contains all the information pertinent to the log entry, such as the logger
            name, log level, file name, line number, and timestamp.

        Returns
        -------
        int
            The length of the formatted header string, which is used to determine the
            amount of indentation needed for multi-line log messages.

        Example
        -------
        >>> record = logging.LogRecord(name="test_logger", level=logging.INFO,
        ... pathname="test.py", lineno=10, msg="Test message", args=None,
        ... exc_info=None)
        >>> formatter = MultiLineFormatter()
        >>> length = formatter.get_header_length(record)
        >>> print(length)
        40  # (for example, depending on the log format)

        """
        return len(
            super().format(
                logging.LogRecord(
                    name=record.name,
                    level=record.levelno,
                    pathname=record.pathname,
                    lineno=record.lineno,
                    msg="",
                    args=(),
                    exc_info=None,
                ),
            ),
        )

    def format(self, record: logging.LogRecord) -> str:
        """Format a log record with added indentation for multi-line messages.

        This method extends the default formatting of a log record by adding indentation
        to any multi-line log messages. The indentation is based on the length of the
        header of the log record, ensuring that all subsequent lines of the message are
        aligned with the header.

        Parameters
        ----------
        record : logging.LogRecord
            The log record to format. This contains all the information necessary to
            produce the log output, including the log level, message, logger name,
            and any other relevant details.

        Returns
        -------
        str
            The formatted log message, with added indentation for multi-line entries.
            The first line contains the standard log header, while subsequent lines
            are indented according to the header's length.

        """
        frame = inspect.currentframe()
        method_name = "unknown"
        max_frames_to_go_up = 8
        frames_up = 0
        while frame is not None and frames_up < max_frames_to_go_up:
            frame = frame.f_back
            frames_up += 1
        if frame is not None:
            method_name = frame.f_code.co_name
        record.name = f"{record.name}/{method_name}"
        indent = " " * self.get_header_length(record)
        head, *trailing = super().format(record).splitlines(keepends=True)
        return head + "".join(indent + line for line in trailing)



"""Module for computing threshold profiles based on relative standard deviations.

This module provides tools to calculate profiles that deviate from the mean profile
by a user-defined multiple of the standard deviation derived from historical data.
The profiles aim to capture variability while maintaining a consistent baseline.
"""

import pandas as pd

from energy_analysis_toolbox.timeseries.profiles.mean_profile import MeanProfile


class RelativeSTDThreshold(MeanProfile):
    """A class which implements a statistical deviation from the mean profile."""

    def __init__(
        self,
        offset_std: float = 3,
        **kwargs,
    ) -> None:
        """Initialize RelativeSTDThreshold.

        Parameters
        ----------
        offset_std : float, optional
            Number of standard deviations VS the computed reference to obtain
            the threshold profile. Default is 3 (profile is 3 standard deviations
            from reference)

        """
        self.offset_std = offset_std
        super().__init__(**kwargs)

    def compute(
        self,
        history: pd.Series,
        time: pd.Timestamp,
        **kwargs,
    ) -> pd.Series:
        """Return a threshold profile.

        The threshold profile is obtained using a user-defined relative variation
        from the mean profile built from history.

        Parameters
        ----------
        history : pd.Series
            Consumption history used to computed the reference
            profile.
        time : pd.Timestamp
            The time at which the profile is of interest. Only
            the information about the date is used in the passed
            timestamp.

        Returns
        -------
        profile : pd.Series
            Profile threshold with same resolution as the history data

        Notes
        -----
        The profile threshold is obtained as the mean profile obtained
        from history + ``tshd`` times the standard deviation profile.

        """
        profile_groups = self.group(history)
        reference = super().compute(history, time, **kwargs)
        offset = self.offset_std * profile_groups.std()
        offset.index = reference.index
        return reference + offset



"""Function to resample a energy series."""

from typing import Literal

import pandas as pd

from energy_analysis_toolbox.timeseries.resample.conservative import volume_to_freq


def to_freq(
    series: "pd.Series[float]",
    freq: str,
    origin: Literal["floor", "ceil"] | None | pd.Timestamp = None,
    last_step_duration: float | None = None,
) -> "pd.Series[float]":
    """Resample an energy series to a given frequency.

    The last step duration of the resampled series is set to the frequency ``freq``.

    Parameters
    ----------
    series : pd.Series[float]
        a series of energy.
    freq : str
        the frequency to resample to.
    origin : {"floor", "ceil", pd.Timestamp}, optional
        the origin of the resampling, by default None.
        See :py:func:`flow_rate_to_freq` for more details.
    last_step_duration : float, optional
        Duration of the last time-step in the ``volume`` series in (s).
        The default is |None| in which case the duration of the former-last
        time-step is used.
        See :py:func:`flow_rate_to_freq` for more details.

    Returns
    -------
    pd.Series[float]
        the resampled energy series.


    .. seealso::

        * :py:func:`energy_analysis_toolbox.timeseries.resample.
          conservative.volume_to_freq`
        * :py:func:`energy_analysis_toolbox.timeseries.resample.
          conservative.volume_conservative`

    """
    if series.empty:
        return series
    return volume_to_freq(series, freq, origin, last_step_duration)



"""Module for filtering historical data based on time-based categories.

This module provides utility functions to subset or filter historical data by
categorizing time-indexed rows using a user-defined classificator. The filters
allow for keeping or removing rows based on matching categories or inclusion
in specified category lists.
"""

from collections.abc import Callable

import pandas as pd


def same_category(
    history: pd.DataFrame,
    date: pd.Timestamp | None = None,
    classificator: Callable | None = None,
) -> pd.DataFrame:
    """Return the subset of history for which the category is the same as the date.

    Parameters
    ----------
    history : pd.DataFrame
        History data to be filtered. It is expected that the data is time-indexed,
        with monotonic-increasing labels.
    date : pd.Timestamp, optional
        Reference for which the category should be the same in history.
        In case ``None`` is passed, the start of the day after the one of the
        last entry in history is used.
    classificator : callable or None, optional
        A function mapping a timestamp to a category. If None, return the
        whole history.

    Returns
    -------
    pd.DataFrame :
        Input history in which only the entries for which the returned category
        is the same as which of ``date`` are returned.

    """
    if history.empty or classificator is None:
        return history
    if date is None:
        date = (history.index[-1].floor("D")) + pd.Timedelta("1D")
    categories = history.index.to_series().apply(classificator)
    ref_category = classificator(date)
    return history.loc[categories == ref_category]


def keep_categories(
    history: pd.DataFrame,
    classificator: Callable | None = None,
    keep: list | None = None,
) -> pd.DataFrame:
    """Return the subset of history for which the category is in the list.

    Parameters
    ----------
    history : pd.DataFrame
        History data to be filtered. It is expected that the data is time-indexed,
        with monotonic-increasing labels.
    classificator : callable or None, optional
        A function mapping a timestamp to a category. If None, return the
        whole history.
    keep : list, optional
        A list of categories representation.
        All rows in ``history`` for which index the ``classificator`` returns
        a value which ``not is in keep`` are dumped from the returned history.

    Returns
    -------
    pd.DataFrame :
        Input history in which only the entries for which the returned category
        is in ``keep`` are returned.

    """
    if history.empty or classificator is None:
        return history
    if keep is None:
        keep = []
    categories = history.index.to_series().apply(classificator)
    mask = [(cat_image in keep) for cat_image in categories.to_numpy()]
    return history.loc[mask]


def remove_categories(
    history: pd.DataFrame,
    classificator: Callable | None = None,
    remove: list | None = None,
) -> pd.DataFrame:
    """Return the subset of history for which the category is the same as the date.

    Parameters
    ----------
    history : pd.DataFrame
        History data to be filtered. It is expected that the data is time-indexed,
        with monotonic-increasing labels.
    classificator : callable or None, optional
        A function mapping a timestamp to a category. If None, return the
        whole history.
    remove : list, optional
        A list of categories representation.
        All rows in ``history`` for which index the ``classificator`` returns
        a value which ``is in remove`` are dumped from the returned history.

    Returns
    -------
    pd.DataFrame :
        Input history in which only the entries for which the returned category
        is different from ones in ``remove`` are returned.

    """
    if history.empty or classificator is None:
        return history
    if remove is None:
        remove = []
    categories = history.index.to_series().apply(classificator)
    mask = [(day_cat not in remove) for day_cat in categories.to_numpy()]
    return history.loc[mask]



"""Test resampling functions for extensive variable flows.
"""

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_index_equal, assert_series_equal

from ..constants import DAY, MINUTE
from ..errors import (
    EATEmptySourceError,
    EATEmptyTargetsError,
    EATInvalidTimestepDurationError,
)
from ..timeseries.resample.conservative import (
    deduce_flows,
    flow_rate_conservative,
    volume_conservative,
)
from .fake.timeseries import example_volume_one_day

# =============================================================================
# --- deduce_flows
# =============================================================================


def test_deduce_flows_nominal():
    """Check that flow-rates are correctly deduced.

    Adapt the input so that the flow rate is always 0.1 m3.s-1.
    """
    start = pd.Timestamp.now()
    s = pd.Timedelta("1s")
    volumes = pd.Series(
        np.array([1.0, 0.5, 2.0, 1.0]),
        index=pd.DatetimeIndex(
            [
                start,
                start + 10 * s,
                start + 15 * s,
                start + 35 * s,
            ],
        ),
    )
    flow_rates = deduce_flows(volumes, last_step_duration=10.0)
    assert np.allclose(flow_rates.values, np.ones(4) * 0.1)
    assert_index_equal(flow_rates.index, volumes.index)


# =============================================================================
# --- flow_rate_conservative
# =============================================================================
def compare_flow_rates(left, right):
    """Factorize flow-rates series comparisons for the tests below."""
    assert_series_equal(
        left, right, check_dtype=False, check_exact=False, check_freq=False,
    )


def test_flow_rate_errors():
    """Check that declared errors are raised."""
    example = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    with pytest.raises(EATEmptySourceError):
        flow_rate_conservative(pd.Series([], dtype="float64"), example.index)
    with pytest.raises(EATEmptyTargetsError):
        flow_rate_conservative(example, pd.DatetimeIndex([]))


def test_flow_rate_out_of_bounds_right():
    """Check the out of bounds resampling on the right."""
    sources = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    target_instants = pd.date_range(
        pd.Timestamp("2022-12-16"), periods=10, freq="12min",
    )
    # default 0 padding
    interp = flow_rate_conservative(sources, target_instants)
    compare_flow_rates(interp, pd.Series(np.zeros(10), index=target_instants))


def test_flow_rate_right_bound_overlap():
    """Check resampling over the right bound.

    Scenario:

        t0                t0+20min
        |    3. X.s-1        |    0. X.s-1                     fr source
        |  3. X.s-1   |   1. X.s-1     |                       fr target
        t0         t0 +15min          (t0+30min)

    """
    sources = pd.Series(
        [3.0],
        pd.date_range(pd.Timestamp("2021-12-15"), periods=1, freq="20min"),
    )
    target_instants = pd.date_range(pd.Timestamp("2021-12-15"), periods=2, freq="15min")
    # implicit last target instant
    interp = flow_rate_conservative(
        sources, target_instants, last_step_duration=20 * MINUTE,
    )
    compare_flow_rates(interp, pd.Series([3.0, 1.0], index=target_instants))
    # explicit last target instant
    interp = flow_rate_conservative(
        sources,
        target_instants,
        last_step_duration=20 * MINUTE,
        last_target_step_duration=30 * MINUTE,
    )
    compare_flow_rates(interp, pd.Series([3.0, 0.5], index=target_instants))


def test_flow_rate_out_of_bounds_left():
    """Check the out of bounds resampling on the left."""
    sources = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    target_instants = pd.date_range(
        pd.Timestamp("2020-12-14"), periods=10, freq="12min",
    )
    # default 0 padding
    interp = flow_rate_conservative(sources, target_instants)
    compare_flow_rates(interp, pd.Series(np.zeros(10), index=target_instants))


def test_flow_rate_identity():
    """Check that the function is identity when source and targets are the same."""
    sources = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    compare_flow_rates(flow_rate_conservative(sources, sources.index), sources)


def test_flow_rate_limit_last_steps():
    """Check that null last timesteps raise errors.

    Checked cases :

        - null last step duration -> Error
        - null last target step duration -> Error
        - null last step and target step durations -> Error
        - negative last step duration -> Error
        - negative last target step duration -> Error

    """
    sources = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    with pytest.raises(EATInvalidTimestepDurationError):
        flow_rate_conservative(sources.iloc[:8], sources.index, last_step_duration=0.0)
    with pytest.raises(EATInvalidTimestepDurationError):
        flow_rate_conservative(sources, sources.index, last_target_step_duration=0.0)
    with pytest.raises(EATInvalidTimestepDurationError):
        flow_rate_conservative(
            sources,
            sources.index,
            last_step_duration=0.0,
            last_target_step_duration=0.0,
        )
    with pytest.raises(EATInvalidTimestepDurationError):
        flow_rate_conservative(sources, sources.index, last_step_duration=-42.0)
    with pytest.raises(EATInvalidTimestepDurationError):
        flow_rate_conservative(sources, sources.index, last_target_step_duration=-42.0)


def test_flow_rate_conservative():
    """Scenario:
    
    .. code-block::

                -30s    t0        t1              t2          t3     t4           t5
        dur               |   60s  |      120s      |   +60s   | 30s |    120s    |
        fr*1e3       0    |   1    |      2         |     3    |  4        |    (0)

        fr*1e3    0   | 0.5   |  1.5   | 2 |   2    |     7/3        |      1     |
            m0     m1      m2        m3  m4      m5               m6           m7

    """
    one_s = pd.Timedelta("1s")
    dur_0 = 60
    dur_1 = 120
    dur_2 = 60  # last step duration inferred from dur_2
    dur_3 = dur_2 / 2
    dur_4 = dur_2 * 2
    t0 = pd.Timestamp("2021-01-06 12:03")
    t1 = t0 + dur_0 * one_s
    t2 = t1 + dur_1 * one_s
    t3 = t2 + dur_2 * one_s
    t4 = t3 + dur_3 * one_s
    t5 = t4 + dur_4 * one_s
    flow_rates = pd.Series(np.arange(1, 5) * 1e-3, pd.DatetimeIndex([t0, t1, t2, t3]))
    target_instants = pd.DatetimeIndex(
        [
            t0 - 3600 * one_s,
            t0 - dur_0 / 2 * one_s,  # m1
            t0 + dur_0 / 2 * one_s,  # m2
            t1 + dur_0 / 2 * one_s,  # m3
            t1 + dur_1 / 2 * one_s,  # m4
            t2,  # m5
            t4,  # m6
            t5,  # m7
        ],
    )
    expected_fr_values = np.array(
        [
            0.0,
            flow_rates.iloc[0] / 2,
            flow_rates.iloc[0:2].mean(),
            flow_rates.iloc[1],
            flow_rates.iloc[1],
            (flow_rates.iloc[2] * 2 + flow_rates.iloc[3] * 1)
            / 3,  # dur_3 is half dur_2
            flow_rates.iloc[3] * 1 / 4,  # dur_4 is 3x dur_3
            0.0,
        ],
    )
    fr_interp = flow_rate_conservative(flow_rates, target_instants)
    compare_flow_rates(fr_interp, pd.Series(expected_fr_values, target_instants))


# =============================================================================
# --- volume_conservative function
# =============================================================================
def compare_volumes(left, right):
    """Factorize volume series comparisons for the tests below."""
    assert_series_equal(
        left, right, check_dtype=False, check_exact=False, check_freq=False,
    )


def test_volume_conservative_errors():
    """Check that declared errors are raised."""
    example = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    with pytest.raises(EATEmptySourceError):
        volume_conservative(pd.Series([], dtype="float64"), example.index)
    with pytest.raises(EATEmptyTargetsError):
        volume_conservative(example, pd.DatetimeIndex([]))


def test_volume_conservative_out_of_bounds_right():
    """Check the out of bounds resampling on the right."""
    sources = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    target_instants = pd.date_range(
        pd.Timestamp("2022-12-15"), periods=10, freq="12min",
    )
    interp = volume_conservative(sources, target_instants)
    compare_volumes(interp, pd.Series(np.zeros(10), index=target_instants))


def test_volume_conservative_out_of_bounds_left():
    """Check the out of bounds resampling on the left."""
    sources = pd.Series(
        np.arange(0, 10),
        pd.date_range(pd.Timestamp("2021-12-15"), periods=10, freq="45min"),
    )
    target_instants = pd.date_range(
        pd.Timestamp("2020-12-15"), periods=10, freq="12min",
    )
    # default 0 padding
    interp = volume_conservative(sources, target_instants)
    compare_volumes(interp, pd.Series(np.zeros(10), index=target_instants))


def test_volume_conservative_identity():
    """Check that the function is identity when source and targets are the same

    Scenario::

        t0       +dt        +2dt        (+3dt)
        |    0    |     50    |    0

    """
    t0 = pd.Timestamp.now()
    dt = pd.Timedelta(seconds=300)
    volumes = pd.Series(
        [0.0, 0.050, 0.0], index=pd.DatetimeIndex([t0, t0 + dt, t0 + 2 * dt]),
    )
    compare_volumes(volumes, volume_conservative(volumes, volumes.index))


def test_volume_conservative_limit_last_steps():
    """Check that limit last timesteps are managed consistently.

    Checked cases :

        - null last target step duration -> Error
        - null last step duration -> Error
        - negative last step duration -> Error
        - negative last target step duration -> Error

    """
    t0 = pd.Timestamp.now()
    dt = pd.Timedelta(seconds=300)
    volumes = pd.Series(
        [0.0, 0.050, 0.05], index=pd.DatetimeIndex([t0, t0 + dt, t0 + 2 * dt]),
    )
    with pytest.raises(EATInvalidTimestepDurationError):
        volume_conservative(volumes, volumes.index, last_target_step_duration=0.0)
    with pytest.raises(EATInvalidTimestepDurationError):
        volume_conservative(volumes, volumes.index, last_step_duration=0.0)
    with pytest.raises(EATInvalidTimestepDurationError):
        volume_conservative(volumes, volumes.index, last_step_duration=-1.0)
    with pytest.raises(EATInvalidTimestepDurationError):
        volume_conservative(volumes, volumes.index, last_target_step_duration=-1.0)


def test_volume_conservative_1():
    """Staggered case.

    Scenario::

        t0       +dt        +2dt        (+3dt)
        |    0    |     50    |    0
             |    25    |    25
           +dt/2      +3dt/2        (+5dt/2)

    """
    t0 = pd.Timestamp.now()
    dt = pd.Timedelta(seconds=300)
    volumes = pd.Series(
        [0.0, 0.050, 0.0], index=pd.DatetimeIndex([t0, t0 + dt, t0 + 2 * dt]),
    )
    volumes_expect = pd.Series(
        [0.025, 0.025], index=pd.DatetimeIndex([t0 + dt / 2, t0 + 3 * dt / 2]),
    )
    compare_volumes(volumes_expect, volume_conservative(volumes, volumes_expect.index))


def test_volume_conservative_2():
    """Subinterval case.

    Scenario::

        t0             +dt                  +2dt            (+3dt)
        |    0          |          60           |    0
        .                     |    20    |   20
                            +4dt/3    +5dt/3   (2dt)
    """
    t0 = pd.Timestamp.now()
    dt = pd.Timedelta(seconds=300)
    volumes = pd.Series(
        [0.0, 0.060, 0.0], index=pd.DatetimeIndex([t0, t0 + dt, t0 + 2 * dt]),
    )
    volumes_expect = pd.Series(
        [0.020, 0.020],
        index=pd.DatetimeIndex([t0 + 4 * dt / 3, t0 + 5 * dt / 3]),
    )
    compare_volumes(volumes_expect, volume_conservative(volumes, volumes_expect.index))


def test_volume_conservative_finer():
    """Resample volume to a regular finer resolution.

    Just check that the total volume is conserved.
    """
    volumes = example_volume_one_day("2020-01-06")
    fine_ixs = pd.date_range(
        start="2020-01-06", end="2020-01-07", inclusive="left", freq="43s",
    )
    fine_volumes = volume_conservative(volumes, fine_ixs)
    assert fine_volumes.sum() == volumes.sum()


def test_volume_conservative_finer_2():
    """Subsampling case.

    Scenario::

        t0             +dt                  +2dt            (+3dt)
        |       40      |          60           |      50
        |  20   |   20  |    30     |     30    |   25 |   25  |

    """
    t0 = pd.Timestamp.now()
    dt = pd.Timedelta(seconds=300)
    volumes = pd.Series(
        [0.040, 0.060, 0.050],
        index=pd.date_range(t0, t0 + 3 * dt, inclusive="left", freq="300s"),
    )
    volumes_expect = pd.Series(
        [0.020, 0.020, 0.030, 0.030, 0.025, 0.025],
        index=pd.date_range(t0, t0 + 3 * dt, inclusive="left", freq="150s"),
    )
    compare_volumes(volumes_expect, volume_conservative(volumes, volumes_expect.index))


def test_volume_conservative_ordinary():
    """Resample volume to an ordinary uneven resolution.

    Scenario::

                 t0.........+10min............//+1h.......//......+1d.................(+1d)
                 |    0 m3     |      3.6 m3  //  |  0m3  //       |     0.864m4......
                 |             |   0.001m3/s  //  |       //       |     1e-5m3/s.....
          x   x             x     x      x                    x       x                (?)
            0        0        0.03   0.6         2.970          0.009       ?

    """
    volumes = pd.Series(
        [0.0, 3.6, 0.0, 0.864],
        index=pd.DatetimeIndex(
            [
                pd.Timestamp("2022-02-22 22:20:22"),
                # 0 m3 here
                pd.Timestamp("2022-02-22 22:30:22"),  # 600s later
                # 3.6 m3 here (1L/s)
                pd.Timestamp("2022-02-22 23:30:22"),  # 3600s later
                # 0 m3 here
                pd.Timestamp("2022-02-23 23:30:22"),  # 1 day later
                # 0.864 m3 here
            ],
        ),
    )
    volumes_expect = pd.Series(
        [0.0, 0.0, 0.030, 0.600, 2.970, 0.009, 0.072],
        index=pd.DatetimeIndex(
            [
                pd.Timestamp("2022-02-22 20:11:13"),
                # 0 m3 here
                pd.Timestamp("2022-02-22 21:14:32"),
                # 0 m3 here
                pd.Timestamp("2022-02-22 22:29:52"),
                # 0.030 m3 here
                pd.Timestamp("2022-02-22 22:30:52"),  # 1 min later, 30s with "draw"
                # 0.600 m3 here
                pd.Timestamp("2022-02-22 22:40:52"),  # 600s later
                # 2.970 m3 here, the rest of the 3.6 m3
                pd.Timestamp("2022-02-23 21:45:22"),
                # 0.009 m3 here
                pd.Timestamp(
                    "2022-02-23 23:45:22",
                ),  # 2h later, 15 min after the new consumption
                # Depends on the duration of last target timestep
                # 0.855 m3 here if forced to 1d
                # 0.072 m3 if let implicit (2h)
            ],
        ),
    )
    volumes_obtained = volume_conservative(volumes, volumes_expect.index)
    compare_volumes(volumes_obtained, volumes_expect)
    # explicit last_target_step_duration
    volumes_obtained = volume_conservative(
        volumes,
        volumes_expect.index,
        last_target_step_duration=DAY,
    )
    volumes_expect.iloc[-1] = 0.855
    compare_volumes(volumes_obtained, volumes_expect)
    # explicit last steps durations
    volumes_obtained = volume_conservative(
        volumes,
        volumes_expect.index,
        last_step_duration=30 * MINUTE,
        last_target_step_duration=DAY,
    )
    volumes_expect.iloc[-2:] = 0.864 / 2
    compare_volumes(volumes_obtained, volumes_expect)


def test_volume_conservative_coarser():
    """Resample volume with a coarser resolution"""
    volumes = example_volume_one_day("2020-01-06")
    coarse_ixs = pd.date_range(start="2020-01-06", end="2020-01-06 12:00", periods=2)
    coarse_volumes = volume_conservative(volumes, coarse_ixs)
    assert coarse_volumes.sum() == pytest.approx(volumes.sum())
    assert coarse_volumes.loc["2020-01-06 12:00":].sum() == pytest.approx(
        volumes.loc["2020-01-06 12:00":].sum(),
    )


def test_volume_conservative_as_integral():
    """Resample volume on a big timestep straddling the series

    The expected result is the sum of all volumes.

    """
    volumes = example_volume_one_day("2020-01-06")
    coarse_ixs = pd.date_range(start="2020-01-05", end="2020-01-07", periods=1)
    coarse_volumes = volume_conservative(
        volumes, coarse_ixs, last_target_step_duration=DAY * 3,
    )
    assert coarse_volumes.loc["2020-01-05"] == pytest.approx(volumes.sum())
    assert coarse_volumes.size == 1



"""Utility functions to compare profiles.
"""

import pandas as pd


def compare_profiles(left, right):
    pd.testing.assert_series_equal(
        left,
        right,
        check_exact=False,
        check_names=False,
        check_freq=False,
    )



"""Convert power timeseries of flows and volumes without breaking conservation laws."""

from typing import Literal

import numpy as np
import pandas as pd

from energy_analysis_toolbox.errors import (
    EATEmptySourceError,
    EATEmptyTargetsError,
    EATInvalidTimestepDurationError,
)
from energy_analysis_toolbox.timeseries.extract_features.basics import (
    index_to_timesteps,
    timestep_durations,
)
from energy_analysis_toolbox.timeseries.resample.index_transformation import (
    index_to_freq,
)
from energy_analysis_toolbox.timeseries.resample.interpolate import piecewise_affine


# =============================================================================
# Resampling with volume conservation
# =============================================================================
def deduce_flows(
    volumes: pd.Series,
    last_step_duration: float | None = None,
) -> pd.Series:
    """Return a timeseries of "flow-rates" from one of an extensive variable.

    Parameters
    ----------
    volumes : pd.Series
        The timeseries of "volumes" (X). In practice, any extensive variable which
        "flows" during the overconsumption of the series.
        The series is assumed to be indexed such that the value at a certain
        index is the volume which flows *until the next index*.
    last_step_duration : float, optional
        Duration of the last time-step in the series in (s).
        The default is |None| meaning that the same duration as the former-last
        one is used.

    Returns
    -------
    pd.Series
        The series of flow-rates in (X.s-1).

    Notes
    -----
    The flow-rate is simply obtained as the ratio between the volume flowing
    during each time-step VS the time-step duration.

    As the values in the series are volumes consumed until the next index, the
    duration to be associated with the last element in the series is ambiguous.
    The chosen value is controlled by ``last_step_duration`` argument.

    """
    durations = timestep_durations(volumes, last_step_duration)
    return volumes / durations


def flow_rate_conservative(
    flow_rates: pd.Series,
    target_instants: pd.DatetimeIndex,
    last_step_duration: float | None = None,
    last_target_step_duration: float | None = None,
) -> pd.Series:
    """Return the series of flow-rate on target instants.

    In what follows, flow-rates and volumes are considered in a broad meaning
    of these terms :

    - volumes represent amounts of an extensive variable which is subjected
      to a conservation law, such as actual volume of water (under certain
      assumptions), or mass, or energy;
    - flow-rates represent flows of this variable, such as volumic flow-rate,
      mass flow-rate or power.

    Parameters
    ----------
    flow_rates : pd.Series
        Timeseries of flow-rates from which to interpolate in (X.s-1) where X
        is any extensive variable.
    target_instants : pd.DatetimeIndex
        Instants at which the flow-rates values have to be returned.
    last_step_duration : float, optional
        Duration of the last time-step in the series in (s).
        The default is |None| in which case the duration of the former-last
        time-step is used.
    last_target_step_duration : float, optional
        Duration of the last time-step in the returned series in (s).
        The default is |None| in which case the duration of the former-last
        time-step is used.
        This duration is used to deduce, if any, what proportion of the volume
        consumption implicitly defined by the ``flow_rates`` series located
        after the last target instant should be attributed to this instant.


    .. important::

        The ``flow_rates`` series is assumed to be indexed such that the value
        at time ``ti`` is the flow-rate during the interval ``[ti, ti+1[`` until
        the next value in the series.

    .. seealso::

        :py:func:`volume_conservative` which resamples the consumed volumes
        on each interval.

    Raises
    ------
    EATEmptySourceError :
        In case ``flow_rates`` is empty.
    EATEmptyTargetsError :
        In case ``target_instants`` is empty.
    EATInvalidTimestepDurationError :
        In case ``last_step_duration <= 0``
    EATInvalidTimestepDurationError :
        In case ``last_target_step_duration <= 0``


    .. note ::

        Uncaught zero-division error will be encountered in case there is a
        null timestep in the source or target values.


    Returns
    -------
    pd.Series :
        Flow-rates values in interpolated at the target instants.


    Notes
    -----
    The flow-rates are interpolated in such a way that the "volume is conserved".
    This means that the volume obtained when integrating the resampled flow-rate
    between two indices in the series should be the same as when computing this
    volume from the initial series.

    The algorithm used by the function is the following :

    - First, the interval durations are computed. The last interval duration
      ambiguity is removed thanks to  ``last_step_duration``. [1.]
    - A series of "volumes" consumed during each timestep is created. This
      is the series of the flow-rate integrals over the time-step in the
      series assuming the ``flow_rates`` series defines a piecewise constant
      function. [2.]
    - The volumes are resampled in a conservative way using
      :py:func:`energy_analysis_toolbox.timeseries.power.volume_conservative`
      function. [3.]
    - The resulting flow-rates are deduced as the ratio between these volumes
      and the interval durations of the target series. The duration of the
      last interval is defined by ``last_target_step_duration``. [4.]

    """
    if flow_rates.empty:
        err = (
            "Resampling an empty flow-rates series to new instants is an invalid "
            "(undefined) operation."
        )
        raise EATEmptySourceError(err)
    if target_instants.empty:
        err = "Target instants must be provided for the series to be resampled."
        raise EATEmptyTargetsError(err)
    durations = timestep_durations(flow_rates, last_step=last_step_duration)  # [1.]
    volumes = flow_rates * durations  # [2.]
    interp_volumes = volume_conservative(
        volumes,
        target_instants,
        last_step_duration=last_step_duration,
        last_target_step_duration=last_target_step_duration,
    )  # [3.]
    target_durations = index_to_timesteps(target_instants, last_target_step_duration)
    interp_flow_rates = interp_volumes / target_durations  # [4.]
    interp_flow_rates.name = flow_rates.name
    interp_flow_rates.index.name = flow_rates.index.name
    return interp_flow_rates


def volume_conservative(
    volumes: pd.Series,
    target_instants: pd.DatetimeIndex,
    last_step_duration: float | None = None,
    last_target_step_duration: float | None = None,
) -> pd.Series:
    """Resample the volume on target instants assuming it is a conservative variable.

    In what follows, flow-rates and volumes are considered in a broad meaning
    of these terms :

    - volumes represent amounts of an extensive variable such as actual
      volume of water, or mass, or energy;
    - flow-rates represent flows of this variable, such as volumic flow-rate,
      mass flow-rate or power.

    The resampling is conservative in the sense that :

    - assuming that ``volumes`` defines a piecewise constant function defined
      on ``[closed, open[`` overconsumption (i.e. constant flow-rate between
      timesteps),
    - the integral of the function defined by the interpolated volume series
      is always equal to which of the function defined by ``volumes`` on the
      same support.


    Parameters
    ----------
    volumes : pd.Series
        The timeseries of volumes in (X). X is the unit of an "extensive"
        variable which can flow along time.
        The series is assumed to be indexed such that the value at a certain
        index is the volume consumed *until the next index*.
    target_instants : pd.DatetimeIndex
        Instants at which the flow-rates values have to be returned.
    last_step_duration : float, optional
        Duration of the last time-step in the ``volume`` series in (s).
        The default is |None| in which case the duration of the former-last
        time-step is used.
        This duration is used to deduce the flow-rate in the last time-step.
    last_target_step_duration : float, optional
        Duration of the last time-step in the returned series in (s).
        The default is |None| in which case the duration of the former-last
        time-step is used.
        This duration is used to deduce, if any, what proportion of the volume
        in the ``volumes`` series located after the last target instant should
        be attributed to this instant.

    Returns
    -------
    pd.Series
        The volumes series sampled on the overconsumption defined in
        ``target_instants``.

    Raises
    ------
    EATEmptySourceError :
        In case ``volumes`` is empty.
    EATEmptyTargetsError :
        In case ``target_instants`` is empty.
    EATInvalidTimestepDurationError :
        In case ``last_step_duration <= 0``.
    EATInvalidTimestepDurationError :
        In case ``last_target_step_duration <= 0``.

    Example
    -------
    This function can resample extensive variables (typically, volumes) which
    consumption is spread along time using non-matching indices. It assumes a
    constant flux in order to ensure the conservation of the integrated variable
    on the new sampling overconsumption.

    >>> volumes = pd.Series(
            np.array([0, 0.1, 0.05, 0.08]),
            index=pd.date_range("2021-12-15", periods=4, freq='6h')
            )
    >>> volumes
    2021-12-15 00:00:00    0.00
    2021-12-15 06:00:00    0.10
    2021-12-15 12:00:00    0.05
    2021-12-15 18:00:00    0.08
    Freq: 6H, dtype: float64
    >>> target_index = pd.date_range("2021-12-15", periods=3, freq='8h')
    >>> volumes_obtained = volume_conservative(volumes, target_index)
    >>> volumes_obtained
    2021-12-15 00:00:00    0.033333
    2021-12-15 08:00:00    0.100000
    2021-12-15 16:00:00    0.096667

    As the timeseries of extensive variables defined a consumption spread on
    an interval, the last step duration is ambiguous. An arbitrary duration
    can be enforced using the ``last_step_duration`` argument.

    >>> volumes = pd.Series(
            np.array([0.1]),
            index=pd.date_range("2021-12-15", periods=1, freq='1d')
            )
    2021-12-15    0.1
    Freq: D, dtype: float64
    >>> volumes
    >>> target_index = pd.date_range("2021-12-15", periods=4, freq='6h')
    >>> volumes_obtained = volume_conservative(volumes, target_index,
                                           last_step_duration=SK.day)
    >>> volumes_obtained
    2021-12-15 00:00:00    0.025
    2021-12-15 06:00:00    0.025
    2021-12-15 12:00:00    0.025
    2021-12-15 18:00:00    0.025
    Freq: 6H, dtype: float64

    This is also true and applicable for the last target step duration.

    >>> target_index = pd.date_range("2021-12-15", periods=1, freq='12h')
    >>> volumes_obtained = volume_conservative(
            volumes, target_index, last_step_duration=SK.day,
            last_target_step_duration=SK.day / 2)
    >>> volumes_obtained
    2021-12-15    0.05
    Freq: 12H, dtype: float64

    Padding the source series with zeros until the desired end of resampling can
    also ease removing the ambiguity on the "operative" part of the input data.

    Notes
    -----
    The volumes are resampled in such a way that for ``ti`` and ``ti+1`` two
    consecutive indices in ``target_instants``, the resampled volume associated
    with ``ti`` is the volume "consumed" during ``[ti, ti+1[``, assuming a
    constant flow-rate during the time-overconsumption defined in ``volumes``.

    Accordingly, the function computes the resampled volumes as follows :

    - Compute the cumulated sum of volumes for each time-step. [1.]
    - Add a virtual timestep at the end of the volume series located
      ``last_step_duration`` after the last sample, with dummy value. [2.]
    - As the ``volumes`` series was indexed in a way that element at time
      ``t`` was the volume which flows between ``t`` and the next timestep
      the cumulated series is shifted by one step on the right and left-padded
      with a zero. This way, element at position ``t`` is the total volume
      which has flowed until time ``t``. [3.]
    - Also add a virtual timestep in the target series so that the cumulated
      sum at the implicit end of the target series is computed as well. [4.1]
    - Interpolate the series of cumulated volumes as a piecewise affine
      function. Target timesteps outside the convex span of the source-indices
      (including the fictive one) are assigned the corresponding border value:

      * 0 before the beginning of the ``volumes`` series
      * The total cumulated volume after the end of the volume series

      This is consistent with the fact that no volume flow is considered
      outside the overconsumption defined in the series. [4.2]
    - Compute elementwise differences of the cumulated volume series, which
      returns the series of volume "consumed" between each timestep and the
      previous one. [5.]
    - Shift the obtained result one time-step backward in order to match the
      volume-series sampling convention. [6.]
    - Discard the last fictive time-step which has no meaning in this convention. [7.]

    As the flow-rate is the time-derivative of the cumulated consumption curve,
    using a piecewise linear interpolation is equivalent to assuming a constant
    flow-rate during all the time-steps of the ``volumes`` series.

    .. note::

        **Why does this function work ?** I.e. why does it conserve the volume ?
        **Short explanation "with hands" :**

        The sampling convention of the ``volumes`` series is just an indirect way
        of defining a piecewise constant flow-rate function.
        The current method works by resampling exactly the integral of the
        flow-rate function, to, then, recover a new volume series (piecewise
        constant flow-rate function). By construction, the integral of the resulting
        piecewise constant flow-rate function on the overconsumption of its support of
        definition matches which of the original one


    .. seealso::

        :py:func:`flow_rate_conservative` which resamples the flow-rate on the
        time-overconsumption in a volume-conservative way.

    """
    if volumes.empty:
        err = (
            "Resampling an empty volumes series to new instants is an "
            "invalid operation."
        )
        raise EATEmptySourceError(err)
    if target_instants.empty:
        err = "Target instants must be provided for the series to be resampled."
        raise EATEmptyTargetsError(err)
    if (
        last_step_duration is not None
        and last_step_duration <= 0
        or last_target_step_duration is not None
        and last_target_step_duration <= 0
    ):
        err = "Last step duration cannot be zero."
        raise EATInvalidTimestepDurationError(err)
    vol_index = volumes.cumsum()  # [1.]
    # the function deals with None last_step_duration values
    durations = timestep_durations(volumes.iloc[-2:], last_step=last_step_duration)
    ghost_right = vol_index.index[-1] + pd.Timedelta(seconds=durations.iloc[-1])  # [2.]
    vol_index.loc[ghost_right] = np.nan  # np.nan to not break dtype
    vol_index = vol_index.shift(1, fill_value=0.0)  # [3.]
    # the function deals with None last_target_step_duration
    target_durations = index_to_timesteps(
        target_instants[-2:],
        last_target_step_duration,
    )
    target_instants = target_instants.insert(
        target_instants.size,  # at the end
        target_instants[-1] + pd.Timedelta(seconds=target_durations[-1]),
    )  # [4.1]
    interp_vol_index = piecewise_affine(vol_index, target_instants)  # [4.2]
    interp_volumes = interp_vol_index.diff().shift(-1).dropna()  # [5.] [6.] [7.]
    interp_volumes.name = volumes.name
    interp_volumes.index.name = volumes.index.name
    return interp_volumes


def volume_to_freq(
    series: "pd.Series[float]",
    freq: str | pd.Timedelta,
    origin: None | Literal["floor", "ceil"] | pd.Timestamp = None,
    last_step_duration: float | None = None,
) -> "pd.Series[float]":
    """Return a series resampled to freq such that the volume is conserved.

    The last step duration of the resampled series is set to the frequency ``freq``.

    Parameters
    ----------
    series : pd.Series
        A series of values of a volume-alike quantity with a DatetimeIndex.
        The sampling period can be constant or not (in this case, use
        ``last_step_duration`` argument).
    freq : str | pd.Timedelta
        the freq to which the series is resampled. Must be a valid
        pandas frequency.
    origin : {None, 'floor, 'ceil', pd.Timestamp}, optional
        What origin should be used for the target resampling range.
        See :py:func:`.index_to_freq` for details.
    last_step_duration : float, optional
        Duration of the last time-step in the ``series`` in (s).
        The default is |None| in which case the duration of the former-last
        time-step is used.
        See :py:func:`.index_to_freq` for details.

    Returns
    -------
    pd.Series
        The resampled series.


    .. seealso::

        * :py:func:`volume_conservative` which resamples the flow-rate on the
          target instants.
        * :py:func:`flow_rate_to_freq` which resamples the volume on the giver
          frequency.

    """
    target_instants = index_to_freq(series.index, freq, origin, last_step_duration)
    return volume_conservative(
        series,
        target_instants,
        last_step_duration=last_step_duration,
        last_target_step_duration=pd.Timedelta(freq).total_seconds(),
    )


def flow_rate_to_freq(
    series: "pd.Series[float]",
    freq: str | pd.Timedelta,
    origin: None | Literal["floor", "ceil"] | pd.Timestamp = None,
    last_step_duration: float | None = None,
) -> "pd.Series[float]":
    """Return a series resampled to freq such that the flow rate is conserved.

    The last step duration of the resampled series is set to the frequency ``freq``.

    Parameters
    ----------
    series : pd.Series
        A series of values of a flow_rate-alike quantity with a DatetimeIndex.
        The sampling period can be constant or not (in this case, use
        ``last_step_duration`` argument).
    freq : str, pd.Timedelta
        the freq to which the series is resampled. Must be a valid
        pandas frequency.
    origin : {None, 'floor, 'ceil', pd.Timestamp}, optional
        What origin should be used for the target resampling range.
        See :py:func:`.index_to_freq` for details.
    last_step_duration : float, optional
        Duration of the last time-step in the ``series`` in (s).
        The default is |None| in which case the duration of the former-last
        time-step is used.
        This duration is used to deduce the volume in the last time-step
        as well as the last index of the resampled series.

    Returns
    -------
    pd.Series
        The resampled series.


    .. seealso::

        * :py:func:`flow_rate_conservative` which resamples the flow-rate on
          the target instants.
        * :py:func:`volume_to_freq` which resamples the volume on the giver frequency.

    """
    target_instants = index_to_freq(series.index, freq, origin, last_step_duration)
    return flow_rate_conservative(
        series,
        target_instants,
        last_step_duration=last_step_duration,
        last_target_step_duration=pd.Timedelta(freq).total_seconds(),
    )



"""Tests for the ``RelativeThreshold``.
"""

import pandas as pd

from energy_analysis_toolbox.timeseries.profiles.thresholds.relative import (
    RelativeThreshold,
)

from .check import compare_profiles
from .fake_data import sinusoid_history


def test_daily_1():
    """Check that 0 relative offset falls back to mean profile"""
    history = sinusoid_history(freq="30min", noise=0, n_days=7)
    profile = RelativeThreshold(period="D", offset_relative=0)
    daily = profile.compute(history, history.index[-1].ceil("D"))
    expected = history.iloc[:48].copy()
    expected.index += 7 * pd.Timedelta("1D")
    compare_profiles(daily, expected)


def test_daily_2():
    """Check with 30% relative offset on periodic time-series."""
    history = sinusoid_history(freq="30min", noise=0, n_days=7)
    profile = RelativeThreshold(period="D", offset_relative=0.3)
    daily = profile.compute(history, history.index[-1].ceil("D"))
    expected = history.iloc[:48].copy() * 1.3
    expected.index += 7 * pd.Timedelta("1D")
    compare_profiles(daily, expected)



"""Test the Thermosensibility module."""

import numpy as np
import pandas as pd
import pytest

from energy_analysis_toolbox.synthetic.thermosensitive_consumption import (
    CategorySynthTSConsumption,
    DateSynthTSConsumption,
)
from energy_analysis_toolbox.thermosensitivity import (
    AutoCategoricalThermoSensitivity,
    CategoricalThermoSensitivity,
    DailyCategoricalThermoSensitivity,
    DayOfWeekCategoricalThermoSensitivity,
    ThermoSensitivity,
)


class TestThermoSensitivity:
    """ThermoSensitivity class tests.

    The tests are based on synthetic data generated with the DateSynthTSConsumption class.

    The tests asserts that, when the noise is low, we can recover the parameters of the model
    as expected. Includes

    - Calibration of the base temperature (heating, cooling, both)
    - Fitting the model (heating, cooling, both)
    - determining the degree days (heating, cooling, both) when set to auto


    """

    base_energy = 100
    ts_heating = 100
    ts_cooling = 100
    tref_heating = 16.5
    tref_cooling = 23
    noise_std = 0.0001
    interseason_mean_temperature = 20
    frequency = "1D"
    data_generated_size = 200
    expected_intercept_factor = 1

    def setup_method(self):
        self.synth_heating_only = DateSynthTSConsumption(
            base_energy=self.base_energy,
            t_ref_heat=self.tref_heating,
            ts_heat=self.ts_heating,
            ts_cool=0,
            noise_std=self.noise_std,
        )
        self.synth_cooling_only = DateSynthTSConsumption(
            base_energy=self.base_energy,
            t_ref_cool=self.tref_cooling,
            ts_heat=0,
            ts_cool=self.ts_cooling,
            noise_std=self.noise_std,
        )
        self.synth_both = DateSynthTSConsumption(
            base_energy=self.base_energy,
            t_ref_heat=self.tref_heating,
            t_ref_cool=self.tref_cooling,
            ts_heat=self.ts_heating,
            ts_cool=self.ts_cooling,
            noise_std=self.noise_std,
        )
        self.expected_intercept = self.base_energy * self.expected_intercept_factor

    def test_fit_only_heating_no_noise(self):
        """Test the ThermoSensitivity class."""
        data = self.synth_heating_only.random_consumption(size=self.data_generated_size)

        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="heating",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )

        ts.fit()
        assert ts.model.params["heating_degree_days"] == pytest.approx(
            self.ts_heating, rel=1e-1,
        )
        assert ts.model.params["Intercept"] == pytest.approx(
            self.expected_intercept, rel=1e-1,
        )
        with pytest.raises(KeyError):
            ts.model.params["cooling_degree_days"]

    def test_fit_only_cooling_no_noise(self):
        """Test the ThermoSensitivity class."""
        data = self.synth_cooling_only.random_consumption(
            size=self.data_generated_size, start="2022-07-01",
        )

        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={"cooling": self.tref_cooling},
            degree_days_computation_method="mean",
            degree_days_type="cooling",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )

        ts.fit()
        assert ts.model.params["cooling_degree_days"] == pytest.approx(
            self.ts_cooling, rel=1e-1,
        )
        assert ts.model.params["Intercept"] == pytest.approx(
            self.expected_intercept, rel=1e-1,
        )
        with pytest.raises(KeyError):
            ts.model.params["heating_degree_days"]

    def test_fit_heating_and_cooling_no_noise(self):
        """Test the ThermoSensitivity class."""
        data = self.synth_both.random_consumption(size=self.data_generated_size)

        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={
                "heating": self.tref_heating,
                "cooling": self.tref_cooling,
            },
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )

        ts.fit()
        assert ts.model.params["heating_degree_days"] == pytest.approx(
            self.ts_heating, rel=1e-1,
        )
        assert ts.model.params["cooling_degree_days"] == pytest.approx(
            self.ts_cooling, rel=1e-1,
        )

    def test_only_heating_no_noise_tref_calibration(self):
        """Test the ThermoSensitivity class."""
        data = self.synth_heating_only.random_consumption(size=self.data_generated_size)

        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="heating",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )

        ts.calibrate_base_temperatures()
        assert ts.degree_days_base_temperature["heating"] == pytest.approx(
            self.tref_heating, rel=1e-1,
        )
        with pytest.raises(KeyError):
            ts.degree_days_base_temperature["cooling"]

    def test_only_cooling_no_noise_tref_calibration(self):
        """Test the ThermoSensitivity class."""
        data = self.synth_cooling_only.random_consumption(
            size=self.data_generated_size, start="2022-07-01",
        )

        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="cooling",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )

        ts.calibrate_base_temperatures()
        assert ts.degree_days_base_temperature["cooling"] == pytest.approx(
            self.tref_cooling, rel=1e-1,
        )
        with pytest.raises(KeyError):
            ts.degree_days_base_temperature["heating"]

    def test_heating_and_cooling_no_noise_tref_calibration(self):
        """Test the ThermoSensitivity class."""
        data = self.synth_both.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )

        ts.calibrate_base_temperatures()
        assert ts.degree_days_base_temperature["heating"] == pytest.approx(
            self.tref_heating, rel=1e-1,
        )
        assert ts.degree_days_base_temperature["cooling"] == pytest.approx(
            self.tref_cooling, rel=1e-1,
        )

    def test_aggregted_data(self):
        data = self.synth_both.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )

        with pytest.raises(ValueError):
            ts.aggregated_data

        ts.degree_days_type = "heating"
        ts._aggregate_data({"heating": self.tref_heating})
        agg_data = ts.aggregated_data
        assert set(agg_data.columns) == {
            "energy",
            "temperature",
            "heating_degree_days",
        }

        ts.degree_days_type = "cooling"
        ts._aggregate_data({"cooling": self.tref_cooling})
        agg_data = ts.aggregated_data
        assert set(agg_data.columns) == {
            "energy",
            "temperature",
            "cooling_degree_days",
        }

        ts.degree_days_type = "both"
        ts._aggregate_data({"heating": self.tref_heating, "cooling": self.tref_cooling})
        agg_data = ts.aggregated_data
        assert set(agg_data.columns) == {
            "energy",
            "temperature",
            "heating_degree_days",
            "cooling_degree_days",
        }

    def test_model(self):
        data = self.synth_heating_only.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="heating",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        with pytest.raises(ValueError):
            ts.model
        ts.fit()
        assert ts.model is not None

    def test_init(self):
        data = self.synth_heating_only.random_consumption(size=self.data_generated_size)

        with pytest.raises(ValueError):
            ThermoSensitivity(
                energy_data=data["energy"],
                temperature_data=data["T"],
                degree_days_base_temperature={"heating": self.tref_heating},
                degree_days_computation_method="mean",
                degree_days_type="NOT_A_TYPE",
                interseason_mean_temperature=self.interseason_mean_temperature,
                frequency="1H",
            )

        with pytest.raises(ValueError):
            ThermoSensitivity(
                energy_data=data["energy"],
                temperature_data=data["T"],
                degree_days_base_temperature={"heating": self.tref_heating},
                degree_days_computation_method="mean",
                degree_days_type="cooling",
                interseason_mean_temperature=self.interseason_mean_temperature,
                frequency="1H",
            )

        with pytest.raises(ValueError):
            ThermoSensitivity(
                energy_data=data["energy"],
                temperature_data=data["T"],
                degree_days_base_temperature={"cooling": self.tref_cooling},
                degree_days_computation_method="mean",
                degree_days_type="heating",
                interseason_mean_temperature=self.interseason_mean_temperature,
                frequency="1H",
            )

    def test_post_init(self):
        data = self.synth_heating_only.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="auto",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        assert ts.degree_days_type == "heating"
        assert ts.predictors == ["heating_degree_days"]

        data = self.synth_cooling_only.random_consumption(
            size=self.data_generated_size, start="2022-07-01",
        )
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="auto",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        assert ts.degree_days_type == "cooling"
        assert ts.predictors == ["cooling_degree_days"]

        data = self.synth_both.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="auto",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        assert ts.degree_days_type == "both"
        assert ts.predictors == ["heating_degree_days", "cooling_degree_days"]

    def test_calculate_degree_days(self):
        data = self.synth_both.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        degree_days = ts._calculate_degree_days(
            {"heating": self.tref_heating, "cooling": self.tref_cooling},
        )
        assert isinstance(degree_days, pd.DataFrame)
        assert degree_days.index.equals(ts.resampled_temperature.index)
        assert set(degree_days.columns) == {
            "heating_degree_days",
            "cooling_degree_days",
        }
        degree_days = ts._calculate_degree_days({"heating": self.tref_heating})
        assert isinstance(degree_days, pd.DataFrame)
        assert degree_days.index.equals(ts.resampled_temperature.index)
        assert set(degree_days.columns) == {"heating_degree_days"}
        degree_days = ts._calculate_degree_days({"cooling": self.tref_cooling})
        assert isinstance(degree_days, pd.DataFrame)
        assert set(degree_days.columns) == {"cooling_degree_days"}
        assert degree_days.index.equals(ts.resampled_temperature.index)

    def test_calibrate_base_temperature(self):
        data = self.synth_both.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        tref = ts.calibrate_base_temperature(
            dd_type="heating",
            t0=12,
            xatol=1e-3,
        )
        assert tref == pytest.approx(self.tref_heating, rel=1e-1)
        tref = ts.calibrate_base_temperature(
            dd_type="cooling",
            t0=30,
            xatol=1e-3,
        )
        assert tref == pytest.approx(self.tref_cooling, rel=1e-1)
        with pytest.raises(ValueError):
            ts.calibrate_base_temperature(
                dd_type="both",
                t0=30,
                xatol=1e-3,
            )

    def test_calibrate_base_temperatures(self):
        data = self.synth_both.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        ts.calibrate_base_temperatures(
            t0_cooling=30,
            t0_heating=12,
            xatol=1e-3,
        )
        assert ts.degree_days_base_temperature["heating"] == pytest.approx(
            self.tref_heating, rel=1e-1,
        )
        assert ts.degree_days_base_temperature["cooling"] == pytest.approx(
            self.tref_cooling, rel=1e-1,
        )

        ts.degree_days_base_temperature = {}
        ts.degree_days_type = "heating"
        ts.calibrate_base_temperatures(
            t0_cooling=30,
            t0_heating=12,
            xatol=1e-3,
        )
        assert ts.degree_days_base_temperature["heating"] == pytest.approx(
            self.tref_heating, rel=1e-1,
        )
        assert "cooling" not in ts.degree_days_base_temperature

        ts.degree_days_base_temperature = {}
        ts.degree_days_type = "cooling"
        ts.calibrate_base_temperatures(
            t0_cooling=30,
            t0_heating=12,
            xatol=1e-3,
        )
        assert ts.degree_days_base_temperature["cooling"] == pytest.approx(
            self.tref_cooling, rel=1e-1,
        )
        assert "heating" not in ts.degree_days_base_temperature

    def test_repr(self):
        data = self.synth_both.random_consumption(size=self.data_generated_size)
        ts = ThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=self.interseason_mean_temperature,
            frequency=self.frequency,
        )
        assert isinstance(repr(ts), str)
        assert "ThermoSensitivity" in repr(ts)
        assert len(repr(ts).splitlines()) == 5
        ts.fit()
        assert len(repr(ts).splitlines()) == 22
        assert "heating_degree_days" in repr(ts)
        assert "cooling_degree_days" in repr(ts)
        assert "Intercept" in repr(ts)


class TestThermoSensitivityWeeks(TestThermoSensitivity):
    frequency = "7D"
    data_generated_size = 7 * 100
    expected_intercept_factor = 7


class TestCategoricalThermoSensitivity:
    parameters = [
        {"base_energy": 100, "ts_heat": 2, "ts_cool": 1, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 1, "ts_cool": 2, "noise_std": 0.0001},
    ]

    tref_heating = 16.5
    tref_cooling = 23

    @staticmethod
    def category_func(t_samples):
        """Categorise the samples in weekday or weekend.

        Parameters
        ----------
        t_samples : pd.Series
            Any time series

        Returns
        -------
        np.array
            The categories

        """
        return np.where(t_samples.index.dayofweek < 5, "weekday", "weekend")

    def setup_method(self):
        """Setup the test class."""
        self.synth = CategorySynthTSConsumption(
            category_func=self.category_func,
            parameters=self.parameters,
            t_ref_cool=self.tref_cooling,
            t_ref_heat=self.tref_heating,
            list_categories=["weekday", "weekend"],
        )

    def test_init(self):
        """Test that the class can be initialised."""
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        data["category"] = self.category_func(data["T"])
        CategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            categories=data["category"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="heating",
            interseason_mean_temperature=20,
        )

    def test_post_init(self):
        """Test the post init method.
        In particular, the capability to detect the degree days type when set to auto.
        """
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        data["category"] = self.category_func(data["T"])

        ts = CategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            categories=data["category"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="auto",
            interseason_mean_temperature=20,
        )
        assert ts.degree_days_type == "both"
        assert ts.predictors == ["heating_degree_days", "cooling_degree_days"]

    def test_fit(self):
        """Test the fit method."""
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        data["category"] = self.category_func(data["T"])

        ts = CategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            categories=data["category"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=20,
        )

        ts.fit()

        model = ts.model
        np.testing.assert_allclose(
            model.params["heating_degree_days:weekend"],
            self.parameters[1]["ts_heat"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["heating_degree_days:weekday"],
            self.parameters[0]["ts_heat"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["cooling_degree_days:weekend"],
            self.parameters[1]["ts_cool"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["cooling_degree_days:weekday"],
            self.parameters[0]["ts_cool"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["Intercept:weekday"],
            self.parameters[0]["base_energy"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["Intercept:weekend"],
            self.parameters[1]["base_energy"],
            rtol=1e-1,
        )


class TestDailyCategoricalThermoSensitivity:
    parameters = [
        {"base_energy": 100, "ts_heat": 2, "ts_cool": 1, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 1, "ts_cool": 2, "noise_std": 0.0001},
    ]

    tref_heating = 16.5
    tref_cooling = 23

    @staticmethod
    def category_func(t_samples):
        """Categorise the samples in weekday or weekend.

        Parameters
        ----------
        t_samples : pd.Series
            Any time series

        Returns
        -------
        np.array
            The categories

        """
        return np.where(t_samples.index.dayofweek < 5, "weekday", "weekend")

    @staticmethod
    def category_func_series(index):
        """Categorise the samples in weekday or weekend.

        Parameters
        ----------
        t_samples : pd.Series
            Any time series

        Returns
        -------
        np.array
            The categories

        """
        return pd.Series(
            np.where(index.dayofweek < 5, "weekday", "weekend"), index=index,
        )

    def setup_method(self):
        """Setup the test class."""
        self.synth = CategorySynthTSConsumption(
            category_func=self.category_func,
            parameters=self.parameters,
            t_ref_cool=self.tref_cooling,
            t_ref_heat=self.tref_heating,
            list_categories=["weekday", "weekend"],
        )

    def test_init(self):
        """Test that the class can be initialised."""
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        DailyCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            categories_func=self.category_func_series,
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="heating",
            interseason_mean_temperature=20,
        )

    def test_post_init(self):
        """Test the post init method.
        In particular, the capability to detect the degree days type when set to auto.
        """
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        ts = DailyCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            categories_func=self.category_func_series,
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="auto",
            interseason_mean_temperature=20,
        )
        assert ts.degree_days_type == "both"
        assert ts.predictors == ["heating_degree_days", "cooling_degree_days"]

    def test_fit(self):
        """Test the fit method."""
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        ts = DailyCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            categories_func=self.category_func_series,
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=20,
        )

        ts.fit()

        model = ts.model
        np.testing.assert_allclose(
            model.params["heating_degree_days:weekend"],
            self.parameters[1]["ts_heat"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["heating_degree_days:weekday"],
            self.parameters[0]["ts_heat"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["cooling_degree_days:weekend"],
            self.parameters[1]["ts_cool"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["cooling_degree_days:weekday"],
            self.parameters[0]["ts_cool"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["Intercept:weekday"],
            self.parameters[0]["base_energy"],
            rtol=1e-1,
        )
        np.testing.assert_allclose(
            model.params["Intercept:weekend"],
            self.parameters[1]["base_energy"],
            rtol=1e-1,
        )


class TestDayOfWeekCategoricalThermoSensitivity:
    parameters = [
        {"base_energy": 100, "ts_heat": 1, "ts_cool": 2, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 2, "ts_cool": 4, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 3, "ts_cool": 6, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 4, "ts_cool": 8, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 5, "ts_cool": 10, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 6, "ts_cool": 12, "noise_std": 0.0001},
        {"base_energy": 100, "ts_heat": 7, "ts_cool": 14, "noise_std": 0.0001},
    ]

    tref_heating = 16.5
    tref_cooling = 23

    @staticmethod
    def category_func(t_samples):
        """Categorise the samples in weekday or weekend.

        Parameters
        ----------
        t_samples : pd.Series
            Any time series

        Returns
        -------
        np.array
            The categories

        """
        mapping = {
            0: "Monday",
            1: "Tuesday",
            2: "Wednesday",
            3: "Thursday",
            4: "Friday",
            5: "Saturday",
            6: "Sunday",
        }
        return t_samples.index.dayofweek.map(mapping)

    def setup_method(self):
        """Setup the test class."""
        self.synth = CategorySynthTSConsumption(
            category_func=self.category_func,
            parameters=self.parameters,
            t_ref_cool=self.tref_cooling,
            t_ref_heat=self.tref_heating,
            list_categories=[
                "Monday",
                "Tuesday",
                "Wednesday",
                "Thursday",
                "Friday",
                "Saturday",
                "Sunday",
            ],
        )

    def test_init(self):
        """Test that the class can be initialised."""
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        DayOfWeekCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="heating",
            interseason_mean_temperature=20,
        )

    def test_post_init(self):
        """Test the post init method.
        In particular, the capability to detect the degree days type when set to auto.
        """
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        ts = DayOfWeekCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="auto",
            interseason_mean_temperature=20,
        )
        assert ts.degree_days_type == "both"
        assert ts.predictors == ["heating_degree_days", "cooling_degree_days"]

    def test_fit(self):
        """Test the fit method."""
        data: pd.DataFrame = self.synth.random_consumption(size=400)
        ts = DayOfWeekCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=20,
        )

        ts.fit()

        model = ts.model
        for idex, dayname in enumerate(
            [
                "Monday",
                "Tuesday",
                "Wednesday",
                "Thursday",
                "Friday",
                "Saturday",
                "Sunday",
            ],
        ):
            np.testing.assert_allclose(
                model.params[f"heating_degree_days:{dayname}"],
                self.parameters[idex]["ts_heat"],
                rtol=1e-1,
            )
            np.testing.assert_allclose(
                model.params[f"cooling_degree_days:{dayname}"],
                self.parameters[idex]["ts_cool"],
                rtol=1e-1,
            )
            np.testing.assert_allclose(
                model.params[f"Intercept:{dayname}"],
                self.parameters[idex]["base_energy"],
                rtol=1e-1,
            )


class TestAutoCategoricalThermoSensitivity:
    parameters = [
        {
            "base_energy": 100,
            "ts_heat": 1,
            "ts_cool": 2,
            "noise_std": 0.00001,
        },  # class 1
        {
            "base_energy": 100,
            "ts_heat": 2,
            "ts_cool": 4,
            "noise_std": 0.00001,
        },  # class 2
        {
            "base_energy": 100,
            "ts_heat": 1,
            "ts_cool": 2,
            "noise_std": 0.00001,
        },  # class 1
        {
            "base_energy": 100,
            "ts_heat": 4,
            "ts_cool": 8,
            "noise_std": 0.00001,
        },  # class 3
        {
            "base_energy": 100,
            "ts_heat": 4,
            "ts_cool": 8,
            "noise_std": 0.00001,
        },  # class 3
        {
            "base_energy": 100,
            "ts_heat": 2,
            "ts_cool": 4,
            "noise_std": 0.00001,
        },  # class 2
        {
            "base_energy": 100,
            "ts_heat": 1,
            "ts_cool": 2,
            "noise_std": 0.00001,
        },  # class 1
    ]

    tref_heating = 16.5
    tref_cooling = 23

    @staticmethod
    def category_func(t_samples):
        """Categorise the samples in weekday or weekend.

        Parameters
        ----------
        t_samples : pd.Series
            Any time series

        Returns
        -------
        np.array
            The categories

        """
        mapping = {
            0: "Monday",
            1: "Tuesday",
            2: "Wednesday",
            3: "Thursday",
            4: "Friday",
            5: "Saturday",
            6: "Sunday",
        }
        return t_samples.index.dayofweek.map(mapping)

    def setup_method(self):
        """Setup the test class."""
        self.synth = CategorySynthTSConsumption(
            category_func=self.category_func,
            parameters=self.parameters,
            t_ref_cool=self.tref_cooling,
            t_ref_heat=self.tref_heating,
            list_categories=[
                "Monday",
                "Tuesday",
                "Wednesday",
                "Thursday",
                "Friday",
                "Saturday",
                "Sunday",
            ],
        )

    def test_init(self):
        """Test that the class can be initialised."""
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        AutoCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="heating",
            interseason_mean_temperature=20,
        )

    def test_post_init(self):
        """Test the post init method.
        In particular, the capability to detect the degree days type when set to auto.
        """
        data: pd.DataFrame = self.synth.random_consumption(size=200)
        ts = AutoCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={"heating": self.tref_heating},
            degree_days_computation_method="mean",
            degree_days_type="auto",
            interseason_mean_temperature=20,
        )
        assert ts.degree_days_type == "both"
        assert ts.predictors == ["heating_degree_days", "cooling_degree_days"]

    def test_fit(self):
        """Test the fit method."""
        data: pd.DataFrame = self.synth.random_consumption(size=600)
        ts = AutoCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={},
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=20,
        )

        ts.fit()

        model = ts.model
        for idex, dayname in enumerate(
            [
                "Monday",
                "Tuesday",
                "Wednesday",
                "Thursday",
                "Friday",
                "Saturday",
                "Sunday",
            ],
        ):
            np.testing.assert_allclose(
                model.params[f"heating_degree_days:{dayname}"],
                self.parameters[idex]["ts_heat"],
                rtol=2e-1,
            )
            np.testing.assert_allclose(
                model.params[f"cooling_degree_days:{dayname}"],
                self.parameters[idex]["ts_cool"],
                rtol=2e-1,
            )
            np.testing.assert_allclose(
                model.params[f"Intercept:{dayname}"],
                self.parameters[idex]["base_energy"],
                rtol=2e-1,
            )

    def test_merge_and_fit(self):
        """Test the fit method."""
        data: pd.DataFrame = self.synth.random_consumption(size=800)
        ts = AutoCategoricalThermoSensitivity(
            energy_data=data["energy"],
            temperature_data=data["T"],
            degree_days_base_temperature={
                "heating": self.tref_heating,
                "cooling": self.tref_cooling,
            },
            degree_days_computation_method="mean",
            degree_days_type="both",
            interseason_mean_temperature=20,
        )

        ts.fit()
        ts.merge_and_fit(significant_level=0.01)

        model = ts.model
        daynames = [
            "Monday-Wednesday-Sunday",
            "Tuesday-Saturday",
            "Thursday-Friday",
        ]
        indexes = [0, 1, 3]
        print(model.params)
        for idex, dayname in zip(indexes, daynames, strict=False):
            np.testing.assert_allclose(
                model.params[f"heating_degree_days:{dayname}"],
                self.parameters[idex]["ts_heat"],
                rtol=2e-1,
            )
            np.testing.assert_allclose(
                model.params[f"cooling_degree_days:{dayname}"],
                self.parameters[idex]["ts_cool"],
                rtol=2e-1,
            )
            np.testing.assert_allclose(
                model.params[f"Intercept:{dayname}"],
                self.parameters[idex]["base_energy"],
                rtol=2e-1,
            )



"""Test the synthetic data generators"""

import numpy as np
import pandas as pd
import pytest

from ..synthetic import (
    DateSynthTSConsumption,
    SynthDDConsumption,
    SynthTSConsumption,
)


class TestSynthDDConsumption:
    base_energry = 1000
    ts_slope = 10
    noise_std = 5
    noise_seed = 42
    expected_columns = ["DD", "energy", "thermosensitive", "base", "residual"]

    @pytest.fixture
    def setup(self) -> SynthDDConsumption:
        # Initialize your SynthDDConsumption object here
        synth_dd_consumption = SynthDDConsumption(
            base_energy=self.base_energry,
            ts_slope=self.ts_slope,
            noise_std=self.noise_std,
        )
        return synth_dd_consumption

    def reset_seed(self, synth):
        # resetting the seed
        synth._rng = np.random.default_rng(seed=self.noise_seed)

    def test_init(self):
        """Test the initialisation"""
        SynthDDConsumption(
            base_energy=self.base_energry,
            ts_slope=self.ts_slope,
            noise_std=self.noise_std,
        )

    def test_random_dds(self, setup: SynthDDConsumption):
        synth_dd_consumption = setup
        size = 45
        start = "2024-04-18"
        dds = synth_dd_consumption.random_dds(size=size, start=start)
        assert isinstance(dds, pd.Series)
        assert len(dds) == size
        assert dds.index[0] == pd.Timestamp(start)
        assert dds.index.freq == "D"

        # test that start can be a pd.Timestamp
        dds = synth_dd_consumption.random_dds(start=pd.Timestamp(start))
        assert dds.index[0] == pd.Timestamp(start)
        # test that start can be a Date
        dds = synth_dd_consumption.random_dds(start=pd.Timestamp(start).date())
        assert dds.index[0] == pd.Timestamp(start)

    def test_random_consumption(self, setup: SynthDDConsumption):
        synth_dd_consumption = setup
        size = 45
        start = "2024-04-18"
        consumption = synth_dd_consumption.random_consumption(size=size, start=start)
        assert isinstance(consumption, pd.DataFrame)
        assert len(consumption) == size

        for column in self.expected_columns:
            assert column in consumption.columns
        assert consumption.index[0] == pd.Timestamp(start)
        assert consumption.index.freq == "D"
        assert all(consumption["base"] == self.base_energry)
        # generate a large sample to test the noise std
        consumption = synth_dd_consumption.random_consumption(size=10_000)
        assert abs(consumption["residual"].std() - self.noise_std) < 0.05  # 1% error

    def test_measures(self, setup: SynthDDConsumption):
        synth_dd_consumption = setup
        measures = synth_dd_consumption.measures()
        self.reset_seed(synth_dd_consumption)
        consumption = synth_dd_consumption.random_consumption()
        pd.testing.assert_frame_equal(measures, consumption)


class TestSynthTSConsumption(TestSynthDDConsumption):
    """Test the synthetic data generators"""

    base_energry = 1000
    ts_heat = 10
    ts_cool = 5
    t_ref_heat = 15
    t_ref_cool = 25
    noise_std = 5
    noise_seed = 42
    expected_columns = [
        "DD_heating",
        "DD_cooling",
        "energy",
        "thermosensitive",
        "base",
        "residual",
    ]

    @pytest.fixture
    def setup(self) -> SynthTSConsumption:
        # Initialize your SynthTSConsumption object here
        synth_ts_consumption = SynthTSConsumption(
            base_energy=self.base_energry,
            ts_heat=self.ts_heat,
            ts_cool=self.ts_cool,
            t_ref_heat=self.t_ref_heat,
            t_ref_cool=self.t_ref_cool,
            noise_std=self.noise_std,
        )
        return synth_ts_consumption

    def reset_seed(self, synth):
        # resetting the seed
        synth._rng = np.random.default_rng(seed=self.noise_seed)
        synth.heating._rng = np.random.default_rng(seed=self.noise_seed)
        synth.cooling._rng = np.random.default_rng(seed=self.noise_seed)

    def test_random_dds(self, setup: SynthTSConsumption):
        synth_dd_consumption = setup
        size = 45
        start = "2024-04-18"
        dds = synth_dd_consumption.random_dds(size=size, start=start)
        assert isinstance(dds, pd.DataFrame)
        assert len(dds) == size
        assert dds.index[0] == pd.Timestamp(start)
        assert dds.index.freq == "D"

        # test that start can be a pd.Timestamp
        dds = synth_dd_consumption.random_dds(start=pd.Timestamp(start))
        assert dds.index[0] == pd.Timestamp(start)
        # test that start can be a Date
        dds = synth_dd_consumption.random_dds(start=pd.Timestamp(start).date())
        assert dds.index[0] == pd.Timestamp(start)


class TestDateTimeSynthTSConsumption(TestSynthTSConsumption):
    @pytest.fixture
    def setup(self) -> DateSynthTSConsumption:
        # Initialize your SynthTSConsumption object here
        synth_ts_consumption = DateSynthTSConsumption(
            base_energy=self.base_energry,
            ts_heat=self.ts_heat,
            ts_cool=self.ts_cool,
            t_ref_heat=self.t_ref_heat,
            t_ref_cool=self.t_ref_cool,
            noise_std=self.noise_std,
        )
        return synth_ts_consumption



"""Run tests and coverages.

This module implements a ``run`` function which can be used to run the library's
test suite from a python interpreter.

Using this function rather than launching using ``pytest`` enables you to be sure
that the tested code is the one from your ``PYTHONPATH`` i.e., the code which
you use when you type ``import energy_analysis_toolbox`` in the calling interpreter.
"""

import os

import pytest


def run():
    """Use pytest to run the test.

    Preferred way to test the library's installation used by the calling python process
    """
    TEST_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
    test_report_filename = os.getcwd() + "/junit.xml"
    coverage_config_filename = (
        os.path.dirname(os.path.dirname(TEST_DIRECTORY)) + "/.coveragerc"
    )
    if not os.path.isfile(coverage_config_filename):
        coverage_config_filename = os.getcwd() + "/.coveragerc"
    if not os.path.isfile(coverage_config_filename):
        coverage_config_filename = ""
    code = pytest.main(
        [
            TEST_DIRECTORY,
            "--junitxml",
            test_report_filename,
            "--cov",
            "energy_analysis_toolbox",
            "--cov-config",
            coverage_config_filename,
            "--cov-report",
            "xml:coverage.xml",
        ],
    )

    if code != pytest.ExitCode(0):
        raise ValueError("Failed test suite")


if __name__ == "__main__":
    run()



from collections.abc import Callable
from typing import Literal

import pandas as pd

from .conservative import (
    flow_rate_to_freq,
    volume_to_freq,
)
from .index_transformation import index_to_freq
from .interpolate import (
    piecewise_affine,
    piecewise_constant,
)

resampling_methods = (
    Literal[
        "piecewise_affine",
        "piecewise_constant",
        "volume_conservative",
        "flow_rate_conservative",
    ]
    | Callable[[pd.Series, pd.DatetimeIndex], pd.Series]
)


def to_freq(
    timeseries: "pd.Series[float]",
    freq: str,
    origin: Literal["floor", "ceil"] | pd.Timestamp | None = None,
    last_step_duration: float | None = None,
    method: resampling_methods = "piecewise_affine",
    **kwargs,
) -> "pd.Series[float]":
    """Return a timeseries resampled at a given frequency.

    Parameters
    ----------
    timeseries : pd.Series
        Series of values of a function of time, indexed using DateTimeIndex.
    freq : str
        Frequency of the resampled series. See :py:func:`pandas.Series.resample`
        for a list of possible values.
    origin : {None, 'floor', 'ceil', pd.Timestamp}, optional
        Origin of the resampling period. see :py:func:`.index_to_freq` for details.
    last_step_duration : {None, float}, optional
        Duration of the last time-step in `timeseries` in (s). See
        :py:func:`.index_to_freq` for details.
    method : str or callable, optional
        Method used to interpolate the values of the resampled series. The accepted
        values are:

        * 'piecewise_affine': uses :py:func:`.piecewise_affine`, assume the values a
          straight line between two points. The default method.
        * 'piecewise_constant': uses :py:func:`.piecewise_constant`, assume the values
          constant until the next point.
        * 'volume_conservative': uses :py:func:`.volume_to_freq`, conserve the quantity
          of the values. Best to use it for energy timeseries.
        * 'flow_rate_conservative': uses :py:func:`.flow_rate_to_freq`, conserve the
          values time the duration between two points. Best to use it for power
          timeseries.

        If a callable is passed, it must take a :py:class:`pandas.Series` as first
        argument and a :py:class:`pandas.DatetimeIndex` as second argument.
        See the interface of :py:func:`piecewise_affine` function.
        The default is 'piecewise_affine'.


    .. important::

        The various methods may manage extrapolation differently, so this situation
        should be avoided or managed with special care.

    Returns
    -------
    new_series : pd.Series
        Values of the series resampled at the given frequency.

    Examples
    --------
    This function can be used to resample timeseries of different physical nature
    with the right method depending on the physical quantity. For example, a
    timeseries of pointwise temperature can be resampled using piecewise affine
    interpolation:

    >>> new_temp = eat.timeseries.resample.to_freq(temperature, '1min',
    ... method='piecewise_affine')

    The same is true for an energy index :

    >>> new_index = eat.timeseries.resample.to_freq(temperature, '1min',
    ... method='piecewise_affine')

    Regarding a quantity that is conserved over time, such as a volume or a flow
    rate, the resampling should be done using a conservative method. For power and
    energy, dedicated functions exists, but this function can also be used:

    >>> new_volume = eat.timeseries.resample.to_freq(volume, '1min',
    ... method='volume_conservative')
    >>> new_flow_rate = eat.timeseries.resample.to_freq(flow_rate, '1min',
    ... method='flow_rate_conservative')

    See more examples of use in :doc:`/user_guide/Resampling_time_series`.


    """
    # Directly apply the method if it is a conservative method for which an
    # integrated method exists
    integrated_methods = {
        "volume_conservative": volume_to_freq,
        "flow_rate_conservative": flow_rate_to_freq,
    }
    try:
        method = integrated_methods[method]
    except KeyError:
        pass
    else:
        return method(
            timeseries,
            freq,
            origin=origin,
            last_step_duration=last_step_duration,
        )
    # Resample
    target_instants = index_to_freq(
        timeseries.index,
        freq,
        origin=origin,
        last_step_duration=last_step_duration,
    )
    # Ensure each method receives only its required parameters
    if method == "piecewise_constant":
        new_series = piecewise_constant(
            timeseries,
            target_instants,
            left_pad=kwargs.get("left_pad"),
        )
    elif method == "piecewise_affine":
        new_series = piecewise_affine(timeseries, target_instants)
    else:
        kwargs["freq"] = freq
        kwargs["origin"] = origin
        kwargs["last_step_duration"] = last_step_duration
        new_series = method(timeseries, target_instants, **kwargs)
    new_series.index.name = timeseries.index.name
    return new_series


def trim_out_of_bounds(
    data: pd.DataFrame | pd.Series,
    resampled_data: pd.DataFrame | pd.Series,
    fill_value: dict[str, any] | None = None,
) -> pd.DataFrame | pd.Series:
    """Fill resampled data with NA outside the boundaries of initial index.

    Parameters
    ----------
    data : pd.DataFrame or pd.Series
        The table of original data which has been resampled.
    resampled_data : pd.DataFrame or pd.Series
        The result of the resampling.
    fill_value : dict, optional
        A dictionary where keys are column names, and values are the placeholders
        used for samples outside the boundaries of the original index.
        If None, defaults to ``{'value': pd.NA}``.

    Returns
    -------
    resampled_data : pd.DataFrame or Series
        The passed resample data with placeholders set **inplace**.

    """
    if fill_value is None:
        fill_value = {"value": pd.NA}
    if resampled_data.index[0] < data.index[0]:
        for col, value in fill_value.items():
            resampled_data.loc[resampled_data.index < data.index[0], col] = value
    if resampled_data.index[-1] > data.index[-1]:
        for col, value in fill_value.items():
            resampled_data.loc[resampled_data.index > data.index[-1], col] = value
    return resampled_data
